/*
 * shadowircd: an advanced internet relay chat daemon
 * 
 * ShadowIRCd is free software. For details on licensing, see the LICENSE file
 * in the root of the source tree.
 * 
 * $Id: newconf.c,v 3.29.2.20 2004/11/20 03:58:07 nenolod Exp $
 */

#include "stdinc.h"

#ifdef HAVE_LIBCRYPTO
#include <openssl/pem.h>
#include <openssl/bio.h>
#include <openssl/rsa.h>
#endif

#include "memory.h"
#include "newconf.h"
#include "tools.h"
#include "ircd_defs.h"
#include "sprintf_irc.h"
#include "common.h"
#include "s_log.h"
#include "s_conf.h"
#include "s_user.h"
#include "send.h"
#include "setup.h"
#include "modules.h"
#include "listener.h"
#include "hostmask.h"
#include "resv.h"
#include "s_serv.h"
#include "event.h"
#include "hash.h"
#include "cluster.h"
#include "ircd.h"
#include "list.h"
#include "irc_getaddrinfo.h"
#include "irc_string.h"
#include "channel.h"

#define CF_TYPE(x) ((x) & CF_MTYPE)

struct TopConf *conf_cur_block;
char	    	*conf_cur_block_name;
char    	*class_name;

static dlink_list conf_items;
static dlink_list col_conf_list = {NULL, NULL, 0};
static dlink_list hub_conf_list = {NULL, NULL, 0};
static dlink_list alconf_items = {NULL, NULL, 0};
static dlink_list leaf_conf_list = {NULL, NULL, 0};

/* static dlink_list yy_conf_list = { NULL, NULL, 0 }; */

/* XXX This _really_ needs to go away */
static struct ConfItem *yy_conf = NULL;
static struct AccessItem *yy_aconf = NULL;
static struct Alias *yy_alconf = NULL;
static struct MatchItem *yy_match_item = NULL;
static struct ClassItem *yy_class = NULL;
static char    *resv_reason;

extern int	filter_add_word(char *);
extern int	filter_del_word(char *);

struct CollectItem {
	dlink_node	node;
	char           *name;
	char           *user;
	char           *host;
	char           *passwd;
	int		port;
	int		flags;
#ifdef HAVE_LIBCRYPTO
	char           *rsa_public_key_file;
	RSA            *rsa_public_key;
#endif
};

struct mode_table flag_table[] = {
	{"remote", OPER_FLAG_REMOTE},
	{"kline", OPER_FLAG_K},
	{"unkline", OPER_FLAG_UNKLINE},
	{"nick_changes", OPER_FLAG_N},
	{"rehash", OPER_FLAG_REHASH},
	{"die", OPER_FLAG_DIE},
	{"global_kill", OPER_FLAG_GLOBAL_KILL},
	{"admin", OPER_FLAG_ADMIN},
	{"hidden_admin", OPER_FLAG_HIDDEN_ADMIN},
	{"xline", OPER_FLAG_X},
	{"auspex", OPER_FLAG_AUSPEX},
	{"routing", OPER_FLAG_ROUTING},
	{"immune", OPER_FLAG_IMMUNE},
	{"override", OPER_FLAG_OVERRIDE},
	{"set_owncloak", OPER_FLAG_OWNCLOAK},
	{"set_anycloak", OPER_FLAG_SETCLOAK},
	{"netadmin", OPER_FLAG_NETADMIN},
	{"techadmin", OPER_FLAG_TECHADMIN},
	{"wants_whois", OPER_FLAG_WANTS_WHOIS},
	{NULL}
};

struct mode_table auth_table[] = {
	{"spoof_notice", CONF_FLAGS_SPOOF_NOTICE},
	{"exceed_limit", CONF_FLAGS_NOLIMIT},
	{"kline_exempt", CONF_FLAGS_EXEMPTKLINE},
	{"no_tilde", CONF_FLAGS_NO_TILDE},
	{"restricted", CONF_FLAGS_RESTRICTED},
	{"need_ident", CONF_FLAGS_NEED_IDENTD},
	{"have_ident", CONF_FLAGS_NEED_IDENTD},
	{"need_password", CONF_FLAGS_NEED_PASSWORD},
	{NULL}
};

struct mode_table connect_table[] = {
	{"autoconn", CONF_FLAGS_ALLOW_AUTO_CONN},
	{"compressed", CONF_FLAGS_COMPRESSED},
	{NULL}
};

void
register_flag(struct mode_table *table, char *flag, int value)
{
	int		i = 0;

	for (i = 0; table[i].name != NULL; i++);

	table[i].name = flag;
	table[i].mode = value;
}

static const char *
conf_strtype(int type)
{
	switch (type & CF_MTYPE) {
		case CF_INT:
		return "integer value";
	case CF_STRING:
		return "unquoted string";
	case CF_YESNO:
		return "yes/no value";
	case CF_QSTRING:
		return "quoted string";
	case CF_TIME:
		return "time/size value";
	default:
		return "unknown type";
	}
}

static void
free_collect_item(struct CollectItem *item)
{
	MyFree(item->name);
	MyFree(item->user);
	MyFree(item->host);
	MyFree(item->passwd);
#ifdef HAVE_LIBCRYPTO
	MyFree(item->rsa_public_key_file);
#endif
	MyFree(item);
}

static void
free_alias_item(struct Alias *item)
{
	MyFree(item->name);
	MyFree(item->command);
	MyFree(item);
}

static void
unhook_hub_leaf_confs(void)
{
	dlink_node     *ptr;
	dlink_node     *next_ptr;
	struct CollectItem *yy_hconf;
	struct CollectItem *yy_lconf;

	DLINK_FOREACH_SAFE(ptr, next_ptr, hub_conf_list.head) {
		yy_hconf = ptr->data;
		dlinkDelete(&yy_hconf->node, &hub_conf_list);
		free_collect_item(yy_hconf);
	}

	DLINK_FOREACH_SAFE(ptr, next_ptr, leaf_conf_list.head) {
		yy_lconf = ptr->data;
		dlinkDelete(&yy_lconf->node, &leaf_conf_list);
		free_collect_item(yy_lconf);
	}
}

static int
		add_top_conf  (const char *name, int (*sfunc) (struct TopConf *), 
		 int (*efunc) (struct TopConf *))
{
	struct TopConf *tc;
	dlink_node     *dn = make_dlink_node();

	tc = MyMalloc(sizeof(struct TopConf));

	DupString(tc->tc_name, name);
	tc->tc_sfunc = sfunc;
	tc->tc_efunc = efunc;

	dlinkAdd(tc, dn, &conf_items);
	return 0;
}

static struct TopConf *
find_top_conf(const char *name)
{
	dlink_node     *d;
	struct TopConf *tc;

	DLINK_FOREACH(d, conf_items.head) {
		tc = d->data;
		if (strcasecmp(tc->tc_name, name) == 0)
			return tc;
	}

	return NULL;
}

#if 0
static int
remove_top_conf(char *name)
{
	struct TopConf *tc;
	dlink_node     *ptr;

	if ((tc = find_top_conf(name)) == NULL)
		return -1;

	if ((ptr = dlinkFind(&conf_items, tc)) == NULL)
		return -1;

	dlinkDestroy(ptr, &conf_items);
	MyFree(tc->tc_name);
	MyFree(tc);

	return 0;
}
#endif

static struct ConfEntry *
find_conf_item(const struct TopConf *top, const char *name)
{
	dlink_node     *d;
	struct ConfEntry *cf;

	DLINK_FOREACH(d, top->tc_items.head) {
		cf = d->data;
		if (strcasecmp(cf->cf_name, name) == 0)
			return cf;
	}

	return NULL;
}

static void
conf_set_serverinfo_rsa_private_key_file(void *data)
{
#ifdef HAVE_LIBCRYPTO
	BIO            *file;

	if (ServerInfo.rsa_private_key) {
		RSA_free(ServerInfo.rsa_private_key);
		ServerInfo.rsa_private_key = NULL;
	}
	if (ServerInfo.rsa_private_key_file) {
		MyFree(ServerInfo.rsa_private_key_file);
		ServerInfo.rsa_private_key_file = NULL;
	}
	DupString(ServerInfo.rsa_private_key_file, (char *)data);

	file = BIO_new_file((char *)data, "r");

	if (file == NULL) {
		conf_report_error
			("Ignoring config file entry rsa_private_key -- file open failed"
			 " (%s)", (char *)data);
		return;
	}
	ServerInfo.rsa_private_key = (RSA *) PEM_read_bio_RSAPrivateKey(file, NULL, 0, NULL);
	if (ServerInfo.rsa_private_key == NULL) {
		conf_report_error
			("Ignoring config file entry rsa_private_key -- couldn't extract key");
		return;
	}
	if (!RSA_check_key(ServerInfo.rsa_private_key)) {
		conf_report_error("Ignoring config file entry rsa_private_key -- invalid key");
		return;
	}
	/* require 2048 bit (256 byte) key */
	if (RSA_size(ServerInfo.rsa_private_key) != 256) {
		conf_report_error("Ignoring config file entry rsa_private_key -- not 2048 bit");
		return;
	}
	BIO_set_close(file, BIO_CLOSE);
	BIO_free(file);
#else
	conf_report_error("Ignoring serverinfo::rsa_private_key -- SSL support not available.");
#endif
}

static void
conf_set_serverinfo_name(void *data)
{
	if (ServerInfo.name == NULL)
		DupString(ServerInfo.name, (char *)data);
}

static void
conf_set_serverinfo_description(void *data)
{
	MyFree(ServerInfo.description);
	DupString(ServerInfo.description, (char *)data);
}

static void
conf_set_serverinfo_sid(void *data)
{
	MyFree(ServerInfo.sid);
	DupString(ServerInfo.sid, (char *)data);
}

static void
conf_set_serverinfo_ssl_certificate_file(void *data)
{
#ifdef HAVE_LIBCRYPTO
	if (ServerInfo.ctx) {

		if (ServerInfo.ssl_certificate_file) {
			MyFree(ServerInfo.ssl_certificate_file);
			ServerInfo.ssl_certificate_file = NULL;
		}
		DupString(ServerInfo.ssl_certificate_file, (char *)data);

		if (!ServerInfo.rsa_private_key_file) {
			conf_report_error
				("Ignoring config file entry ssl_certificate -- no rsa_private_key");
			return;
		}
		if (SSL_CTX_use_certificate_file(ServerInfo.ctx,
					    ServerInfo.ssl_certificate_file,
						 SSL_FILETYPE_PEM) <= 0) {
			return;
		}
		if (SSL_CTX_use_PrivateKey_file(ServerInfo.ctx,
					    ServerInfo.rsa_private_key_file,
						SSL_FILETYPE_PEM) <= 0) {
			return;
		}
		if (!SSL_CTX_check_private_key(ServerInfo.ctx)) {
			conf_report_error
				("RSA private key doesn't match the SSL certificate public key!");
			return;
		}
	}
#endif
}

static void
conf_set_serverinfo_ssl_ca_certificate_file(void *data)
{
#ifdef HAVE_LIBCRYPTO
	if (ServerInfo.ctx) {

		if (ServerInfo.ssl_ca_certificate_file) {
			MyFree(ServerInfo.ssl_ca_certificate_file);
			ServerInfo.ssl_ca_certificate_file = NULL;
		}
		DupString(ServerInfo.ssl_ca_certificate_file, (char *)data);

		if (ServerInfo.ssl_ca_certificate_file != NULL &&
		    SSL_CTX_load_verify_locations(ServerInfo.ctx,
			   ServerInfo.ssl_ca_certificate_file, NULL) != 1) {
			conf_report_error(strncat
					  ("Error using config file entry ssl_ca_certificate -- ",
				    ERR_error_string(ERR_get_error(), NULL),
			  strlen(ERR_error_string(ERR_get_error(), NULL))));
			return;
		}
	}
#endif
}

static void
conf_set_serverinfo_vhost(void *data)
{
	struct addrinfo	hints, *res;

	memset(&hints, 0, sizeof(hints));

	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE | AI_NUMERICHOST;

	if (irc_getaddrinfo((char *)data, NULL, &hints, &res))
		ilog(L_ERROR, "Invalid netmask for server vhost(%s)", (char *)data);
	else {
		ServerInfo.specific_ipv4_vhost = 1;

		assert(res != NULL);

		memcpy(&ServerInfo.ip, res->ai_addr, res->ai_addrlen);
		ServerInfo.ip.ss.ss_family = res->ai_family;
		ServerInfo.ip.ss_len = res->ai_addrlen;
		irc_freeaddrinfo(res);

		ServerInfo.specific_ipv4_vhost = 1;
	}
}

static void
conf_set_serverinfo_vhost6(void *data)
{
#ifdef IPV6
	struct addrinfo	hints, *res;

	memset(&hints, 0, sizeof(hints));

	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE | AI_NUMERICHOST;

	if (irc_getaddrinfo((char *)data, NULL, &hints, &res))
		ilog(L_ERROR, "Invalid netmask for server vhost6(%s)", (char *)data);
	else {
		assert(res != NULL);

		memcpy(&ServerInfo.ip6, res->ai_addr, res->ai_addrlen);
		ServerInfo.ip6.ss.ss_family = res->ai_family;
		ServerInfo.ip6.ss_len = res->ai_addrlen;
		irc_freeaddrinfo(res);

		ServerInfo.specific_ipv6_vhost = 1;
	}
#else
	conf_report_error("Warning -- ignoring serverinfo::vhost6 -- IPv6 support not available.");
#endif
}

static void
conf_set_serverinfo_max_clients(void *data)
{
	int		max_clients = *(int *)data;

	ServerInfo.max_clients = max_clients;
}

static void
conf_set_serverinfo_hub(void *data)
{
	int		hub = *(int *)data;

	ServerInfo.hub = hub;
}

static void
conf_set_modules_module(void *data)
{
#ifndef STATIC_MODULES
	char           *m_bn;

	m_bn = basename((char *)data);

	if (findmodule_byname(m_bn) == -1)
		load_one_module((char *)data, 0);

	MyFree(m_bn);
#else
			conf_report_error("Ignoring modules::module -- loadable module support not present.");
#endif
}

static void
conf_set_modules_path(void *data)
{
#ifndef STATIC_MODULES
	mod_add_path((char *)data);
#else
	conf_report_error("Ignoring modules::path -- loadable module support net present.");
#endif
}

static void
conf_set_admin_name(void *data)
{
	MyFree(AdminInfo.name);
	DupString(AdminInfo.name, (char *)data);
}

static void
conf_set_admin_email(void *data)
{
	MyFree(AdminInfo.email);
	DupString(AdminInfo.email, (char *)data);
}

static void
conf_set_admin_description(void *data)
{
	MyFree(AdminInfo.description);
	DupString(AdminInfo.description, (char *)data);
}

struct mode_table log_levels[] = {
	{"l_crit", L_CRIT},
	{"l_error", L_ERROR},
	{"l_warn", L_WARN},
	{"l_notice", L_NOTICE},
	{"l_trace", L_TRACE},
	{"l_info", L_INFO},
	{"l_debug", L_DEBUG},
	{NULL, 0}
};

static int
find_umode(struct mode_table *tab, char *name)
{
	int		i;

	for (i = 0; tab[i].name; i++) {
		if (strcmp(tab[i].name, name) == 0)
			return tab[i].mode;
	}

	return 0;
}

static void
set_modes_from_table(int *modes, const char *whatis, struct mode_table *tab, conf_parm_t * args)
{
	for (; args; args = args->next) {
		int		mode;

		if ((args->type & CF_MTYPE) != CF_STRING) {
			conf_report_error("Warning -- %s is not a string; ignoring.", whatis);
			continue;
		}
		mode = find_umode(tab, args->v.string);

		if (!mode) {
			conf_report_error("Warning -- unknown %s %s.", whatis, args->v.string);
			continue;
		}
		*modes |= mode;
	}
}

static void
conf_set_logging_log_level(void *data)
{
	set_log_level(find_umode(log_levels, (char *)data));
}

static void
conf_set_logging_fuserlog(void *data)
{
	strlcpy(ConfigLoggingEntry.userlog, (char *)data, sizeof(ConfigLoggingEntry.userlog));
}

static void
conf_set_logging_ffailed_operlog(void *data)
{
	strlcpy(ConfigLoggingEntry.failed_operlog, (char *)data,
		sizeof(ConfigLoggingEntry.failed_operlog));
}

static void
conf_set_logging_foperlog(void *data)
{
	strlcpy(ConfigLoggingEntry.operlog, (char *)data, sizeof(ConfigLoggingEntry.operlog));
}

static void
conf_set_logging_use_logging(void *data)
{
	int		yesno = *(int *)data;
	ConfigLoggingEntry.use_logging = yesno;
}

static int
conf_begin_oper(struct TopConf *tc)
{
	yy_conf = MyMalloc(sizeof(struct ConfItem));
	yy_aconf = MyMalloc(sizeof(struct AccessItem));

	return 0;
}

static int
conf_end_oper(struct TopConf *tc)
{
	struct CollectItem *yy_tmp;
	dlink_node     *ptr, *next_ptr;

	if (conf_cur_block_name != NULL) {
		MyFree(yy_conf->name);
		DupString(yy_conf->name, conf_cur_block_name);
	}
	if (EmptyString(yy_conf->name)) {
		conf_report_error("Ignoring operator block -- missing name");
		return 0;
	}
	DLINK_FOREACH_SAFE(ptr, next_ptr, col_conf_list.head) {
		struct AccessItem *new_aconf;
		struct ConfItem *new_conf;
		yy_tmp = ptr->data;

		new_conf = make_conf_item(OPER_TYPE);
		new_aconf = (struct AccessItem *)map_to_conf(new_conf);

		DupString(new_conf->name, yy_conf->name);

		conf_add_class_to_conf(new_conf, class_name);
		new_aconf->port = yy_aconf->port;
		new_aconf->flags = yy_aconf->flags;

		if (yy_aconf->usermodes)
			DupString(new_aconf->usermodes, yy_aconf->usermodes);

		if (yy_aconf->passwd)
			DupString(new_aconf->passwd, yy_aconf->passwd);

#ifdef HAVE_LIBCRYPTO
		if (yy_aconf->rsa_public_key_file != NULL) {
			BIO            *file;

			DupString(new_aconf->rsa_public_key_file, yy_aconf->rsa_public_key_file);

			file = BIO_new_file(yy_aconf->rsa_public_key_file, "r");
			new_aconf->rsa_public_key = (RSA *) PEM_read_bio_RSA_PUBKEY(file,
							     NULL, 0, NULL);
			BIO_set_close(file, BIO_CLOSE);
			BIO_free(file);
		}
#endif

		/* attempted fix 2 */
		if (yy_tmp->user != NULL) {
			DupString(new_aconf->user, yy_tmp->user);
			collapse(new_aconf->user);
		} else
			DupString(new_aconf->user, "*");

		if (yy_tmp->host != NULL) {
			DupString(new_aconf->host, yy_tmp->host);
			collapse(new_aconf->host);
		} else
			DupString(new_aconf->host, "*");

#ifdef HAVE_LIBCRYPTO
		if (yy_tmp->name && (yy_tmp->passwd || yy_aconf->rsa_public_key) && yy_tmp->host)
#else
		if (yy_tmp->name && yy_tmp->passwd && yy_tmp->host)
#endif
		{
			conf_add_class_to_conf(new_conf, class_name);
			if (yy_tmp->name != NULL)
				DupString(new_conf->name, yy_tmp->name);
		}
		dlinkDelete(&yy_tmp->node, &col_conf_list);
		free_collect_item(yy_tmp);
	}

	yy_conf = NULL;
	yy_aconf = NULL;

	MyFree(class_name);
	class_name = NULL;

	return 0;
}

static void
conf_set_oper_flags(void *data)
{
	conf_parm_t    *args = data;

	set_modes_from_table(&yy_aconf->port, "flag", flag_table, args);
}

static void
conf_set_oper_name(void *data)
{
	MyFree(yy_conf->name);

	if ((strlen((const char *)data)) > OPERNICKLEN)
		((char *)data)[OPERNICKLEN] = '\0';

	DupString(yy_conf->name, (char *)data);
}

static void
conf_set_oper_umodes(void *data)
{
	MyFree(yy_aconf->usermodes);
	DupString(yy_aconf->usermodes, (char *)data);
}

static void
conf_set_oper_user(void *data)
{
	struct CollectItem *yy_tmp;

	yy_tmp = (struct CollectItem *)MyMalloc(sizeof(struct CollectItem));

	DupString(yy_tmp->host, (char *)data);
	split_user_host(yy_tmp->host, &yy_tmp->user, &yy_tmp->host);

	dlinkAdd(yy_tmp, make_dlink_node(), &col_conf_list);
}

static void
conf_set_oper_password(void *data)
{
	if (yy_aconf->passwd)
		memset(yy_aconf->passwd, 0, strlen(yy_aconf->passwd));

	MyFree(yy_aconf->passwd);
	DupString(yy_aconf->passwd, (char *)data);
}

static void
conf_set_oper_rsa_public_key_file(void *data)
{
#ifdef HAVE_LIBCRYPTO
	BIO            *file;

	if (yy_aconf->rsa_public_key != NULL) {
		RSA_free(yy_aconf->rsa_public_key);
		yy_aconf->rsa_public_key = NULL;
	}
	if (yy_aconf->rsa_public_key_file != NULL) {
		MyFree(yy_aconf->rsa_public_key_file);
		yy_aconf->rsa_public_key_file = NULL;
	}
	DupString(yy_aconf->rsa_public_key_file, (char *)data);
	file = BIO_new_file((char *)data, "r");

	if (file == NULL) {
		conf_report_error("Ignoring rsa_public_key_file -- file doesn't exist");
		return;
	}
	yy_aconf->rsa_public_key = (RSA *) PEM_read_bio_RSA_PUBKEY(file, NULL, 0, NULL);

	if (yy_aconf->rsa_public_key == NULL) {
		conf_report_error("Ignoring rsa_public_key_file -- Key invalid; check key syntax.");
		return;
	}
	BIO_set_close(file, BIO_CLOSE);
	BIO_free(file);
#else
	conf_report_error("Warning -- ignoring rsa_public_key_file (OpenSSL support not available");
#endif
}

static void
conf_set_oper_encrypted(void *data)
{
	int		yesno = *(unsigned int *)data;

	if (yesno)
		yy_aconf->flags |= CONF_FLAGS_ENCRYPTED;
	else
		yy_aconf->flags &= ~CONF_FLAGS_ENCRYPTED;
}

static void
conf_set_oper_class(void *data)
{
	MyFree(class_name);
	DupString(class_name, (char *)data);
}

static int
conf_begin_class(struct TopConf *tc)
{
	yy_conf = make_conf_item(CLASS_TYPE);
	yy_class = (struct ClassItem *)map_to_conf(yy_conf);

	return 0;
}

static int
conf_end_class(struct TopConf *tc)
{
	if (conf_cur_block_name != NULL) {
		MyFree(yy_conf->name);
		DupString(yy_conf->name, conf_cur_block_name);
	}
	if ((yy_conf != NULL) && (yy_conf->name == NULL)) {
		delete_conf_item(yy_conf);
		yy_conf = NULL;
		yy_class = NULL;
	}
	return 0;
}

static void
conf_set_class_name(void *data)
{
	MyFree(yy_conf->name);
	DupString(yy_conf->name, (char *)data);
}

static void
conf_set_class_ping_time(void *data)
{
	PingFreq(yy_class) = *(unsigned int *)data;
}

static void
conf_set_class_number_per_ip(void *data)
{
	MaxLocal(yy_class) = *(unsigned int *)data;
}

static void
conf_set_class_number_per_ip_global(void *data)
{
	yy_class->max_global = *(unsigned int *)data;
}

static void
conf_set_class_number_per_ident(void *data)
{
	yy_class->max_ident = *(unsigned int *)data;
}

static void
conf_set_class_connectfreq(void *data)
{
	yy_class->con_freq = *(unsigned int *)data;
}

static void
conf_set_class_max_number(void *data)
{
	yy_class->max_total = *(unsigned int *)data;
}

static void
conf_set_class_sendq(void *data)
{
	yy_class->max_sendq = *(unsigned int *)data;
}

static char    *listener_address;

static int
conf_begin_listen(struct TopConf *tc)
{
	listener_address = NULL;
	return 0;
}

static int
conf_end_listen(struct TopConf *tc)
{
	MyFree(listener_address);
	listener_address = NULL;
	return 0;
}

static void
conf_set_listen_port(void *data)
{
	conf_parm_t    *args = data;

	for (; args; args = args->next) {
		if ((args->type & CF_MTYPE) != CF_INT) {
			conf_report_error("listener::port argument is not an integer "
					  "-- ignoring.");
			continue;
		}
		add_listener(args->v.number, listener_address, 0);
	}
}

static void
conf_set_listen_sslport(void *data)
{
	conf_parm_t    *args = data;

	for (; args; args = args->next) {
		if ((args->type & CF_MTYPE) != CF_INT) {
			conf_report_error("listener::sslport argument is not an integer "
					  "-- ignoring.");
			continue;
		}
		add_listener(args->v.number, listener_address, 1);
	}
}

static void
conf_set_listen_address(void *data)
{
	MyFree(listener_address);
	DupString(listener_address, data);
}

static int
conf_begin_auth(struct TopConf *tc)
{
	yy_conf = make_conf_item(CLIENT_TYPE);
	yy_aconf = map_to_conf(yy_conf);
	yy_aconf->flags |= CONF_FLAGS_SPOOF_NOTICE;	/* default to
							 * spoof_notice=yes: */

	return 0;
}

static int
conf_end_auth(struct TopConf *tc)
{
	struct CollectItem *yy_tmp;
	dlink_node     *ptr;
	dlink_node     *next_ptr;

	conf_add_class_to_conf(yy_conf, class_name);
	add_conf_by_address(CONF_CLIENT, yy_aconf);

	/* copy over settings from first struct */
	DLINK_FOREACH_SAFE(ptr, next_ptr, col_conf_list.head) {
		struct AccessItem *new_aconf;
		struct ConfItem *new_conf;

		new_conf = make_conf_item(CLIENT_TYPE);
		new_aconf = (struct AccessItem *)map_to_conf(new_conf);

		yy_tmp = ptr->data;

		if (yy_aconf->passwd != NULL)
			DupString(new_aconf->passwd, yy_aconf->passwd);
		if (yy_conf->name != NULL)
			DupString(new_conf->name, yy_conf->name);

		if (yy_aconf->passwd != NULL)
			DupString(new_aconf->passwd, yy_aconf->passwd);

		new_aconf->flags = yy_aconf->flags;
		new_aconf->port = yy_aconf->port;

		if (yy_tmp->user != NULL) {
			DupString(new_aconf->user, yy_tmp->user);
			collapse(new_aconf->user);
		} else
			DupString(new_aconf->user, "*");

		if (yy_tmp->host != NULL) {
			DupString(new_aconf->host, yy_tmp->host);
			collapse(new_aconf->host);
		} else
			DupString(new_aconf->host, "*");

		conf_add_class_to_conf(new_conf, class_name);
		add_conf_by_address(CONF_CLIENT, new_aconf);
		dlinkDelete(&yy_tmp->node, &col_conf_list);
		free_collect_item(yy_tmp);
	}

	MyFree(class_name);
	class_name = NULL;
	yy_conf = NULL;
	yy_aconf = NULL;

	return 0;
}

static void
conf_set_auth_user(void *data)
{
	struct CollectItem *yy_tmp;

	if (yy_aconf->user == NULL) {
		if (data != NULL) {
			DupString(yy_aconf->host, (char *)data);
			split_user_host(yy_aconf->host, &yy_aconf->user, &yy_aconf->host);
		}
	} else {
		yy_tmp = (struct CollectItem *)MyMalloc(sizeof(struct CollectItem));
		if (data != NULL) {
			DupString(yy_tmp->host, (char *)data);
			split_user_host(yy_tmp->host, &yy_tmp->user, &yy_tmp->host);
		}
		dlinkAdd(yy_tmp, &yy_tmp->node, &col_conf_list);
	}
}

static void
conf_set_auth_passwd(void *data)
{
	if (yy_aconf->passwd)
		memset(yy_aconf->passwd, 0, strlen(yy_aconf->passwd));
	MyFree(yy_aconf->passwd);
	DupString(yy_aconf->passwd, data);
}

static void
conf_set_auth_encrypted(void *data)
{
	int		yesno = *(unsigned int *)data;

	if (yesno)
		yy_aconf->flags |= CONF_FLAGS_ENCRYPTED;
	else
		yy_aconf->flags &= ~CONF_FLAGS_ENCRYPTED;
}

static void
conf_set_auth_spoof_notice(void *data)
{
	int		yesno = *(unsigned int *)data;

	if (yesno)
		yy_aconf->flags |= CONF_FLAGS_SPOOF_NOTICE;
	else
		yy_aconf->flags &= ~CONF_FLAGS_SPOOF_NOTICE;
}

static void
conf_set_auth_spoof(void *data)
{
	if (strlen(data) > HOSTLEN) {
		conf_report_error("Warning -- spoof host length invalid.");
		return;
	}
	MyFree(yy_conf->name);
	DupString(yy_conf->name, data);
	yy_aconf->flags |= CONF_FLAGS_SPOOF_IP;
}

static void
conf_set_auth_flags(void *data)
{
	conf_parm_t    *args = data;

	set_modes_from_table((int *)&yy_aconf->flags, "flag", auth_table, args);
}

static void
conf_set_auth_exceed_limit(void *data)
{
	int		yesno = *(unsigned int *)data;

	if (yesno)
		yy_aconf->flags |= CONF_FLAGS_NOLIMIT;
	else
		yy_aconf->flags &= ~CONF_FLAGS_NOLIMIT;
}

static void
conf_set_auth_is_restricted(void *data)
{
	int		yesno = *(unsigned int *)data;

	if (yesno)
		yy_aconf->flags |= CONF_FLAGS_RESTRICTED;
	else
		yy_aconf->flags &= ~CONF_FLAGS_RESTRICTED;
}

static void
conf_set_auth_kline_exempt(void *data)
{
	int		yesno = *(unsigned int *)data;

	if (yesno)
		yy_aconf->flags |= CONF_FLAGS_EXEMPTKLINE;
	else
		yy_aconf->flags &= ~CONF_FLAGS_EXEMPTKLINE;
}

static void
conf_set_auth_need_ident(void *data)
{
	int		yesno = *(unsigned int *)data;

	if (yesno)
		yy_aconf->flags |= CONF_FLAGS_NEED_IDENTD;
	else
		yy_aconf->flags &= ~CONF_FLAGS_NEED_IDENTD;
}

static void
conf_set_auth_no_tilde(void *data)
{
	int		yesno = *(unsigned int *)data;

	if (yesno)
		yy_aconf->flags |= CONF_FLAGS_NO_TILDE;
	else
		yy_aconf->flags &= ~CONF_FLAGS_NO_TILDE;
}

static void
conf_set_auth_redir_serv(void *data)
{
	yy_aconf->flags |= CONF_FLAGS_REDIR;
	MyFree(yy_conf->name);
	DupString(yy_conf->name, data);
}

static void
conf_set_auth_redir_port(void *data)
{
	int		port = *(unsigned int *)data;

	yy_aconf->flags |= CONF_FLAGS_REDIR;
	yy_aconf->port = port;
}

static void
conf_set_auth_class(void *data)
{
	MyFree(class_name);
	DupString(class_name, data);
}

static int
conf_begin_resv(struct TopConf *tc)
{
	resv_reason = NULL;
	return 0;
}

static int
conf_end_resv(struct TopConf *tc)
{
	MyFree(resv_reason);
	resv_reason = NULL;
	return 0;
}

static void
conf_set_resv_channel(void *data)
{
	static char	noreason[] = "No Reason";

	if (IsChanPrefix(*(char *)data)) {
		create_channel_resv((char *)data, resv_reason != NULL ? resv_reason : noreason, 1);
	} else {
		conf_report_error
			("Warning -- channel '%s' in resv is not a valid channel name.",
			 (char *)data);
	}
}

static void
conf_set_resv_reason(void *data)
{
	MyFree(resv_reason);
	DupString(resv_reason, data);
}

static void
conf_set_resv_nick(void *data)
{
	char		noreason  [] = "No Reason";
	create_nick_resv((char *)data, resv_reason != NULL ? resv_reason : noreason, 1);
}

static int
conf_begin_connect(struct TopConf *tc)
{
	yy_conf = make_conf_item(SERVER_TYPE);
	yy_aconf = (struct AccessItem *)map_to_conf(yy_conf);
	yy_aconf->passwd = NULL;

	yy_aconf->port = PORTNUM;

	return 0;
}

static int
conf_end_connect(struct TopConf *tc)
{
	struct CollectItem *yy_hconf = NULL;
	struct CollectItem *yy_lconf = NULL;
	dlink_node     *ptr;
	dlink_node     *next_ptr;

	if (conf_cur_block_name != NULL) {
		MyFree(yy_conf->name);
		DupString(yy_conf->name, conf_cur_block_name);
	}
#ifdef HAVE_LIBCRYPTO
	if (yy_aconf->host &&
	    ((yy_aconf->passwd && yy_aconf->spasswd) ||
	     (yy_aconf->rsa_public_key && IsConfCryptLink(yy_aconf))))
#else				/* !HAVE_LIBCRYPTO */
	if (yy_aconf->host && !IsConfCryptLink(yy_aconf) && yy_aconf->passwd && yy_aconf->spasswd)
#endif				/* !HAVE_LIBCRYPTO */
	{
		if (conf_add_server(yy_conf, scount, class_name) >= 0) {
			++scount;
		} else {
			delete_conf_item(yy_conf);
			yy_conf = NULL;
			yy_aconf = NULL;
		}
	} else {
		/*
		 * Even if yy_conf ->name is NULL should still unhook any
		 * hub/leaf confs still pending
		 */
		unhook_hub_leaf_confs();
		if (yy_conf->name != NULL) {
#ifndef HAVE_LIBCRYPTO
			if (IsConfCryptLink(yy_aconf))
				conf_report_error("Ignoring connect block -- no OpenSSL support");
#else
			if (IsConfCryptLink(yy_aconf) && !yy_aconf->rsa_public_key)
				conf_report_error("Ignoring connect block -- missing key");
#endif
			if (yy_aconf->host == NULL)
				conf_report_error("Ignoring connect block -- missing host");
			else if (!IsConfCryptLink(yy_aconf) &&
				 (!yy_aconf->passwd || !yy_aconf->spasswd))
				conf_report_error("Ignoring connect block -- missing password");
		}
		yy_aconf = NULL;
		yy_conf = NULL;
	}

	/*
	 * yy_conf is still pointing at the server that is having a connect
	 * block built for it. This means, y_aconf->name points to the actual
	 * irc name this server will be known as. Now this new server has a
	 * set or even just one hub_mask (or leaf_mask) given in the link
	 * list at yy_hconf. Fill in the HUB confs from this link list now.
	 */
	DLINK_FOREACH_SAFE(ptr, next_ptr, hub_conf_list.head) {
		struct ConfItem *new_hub_conf;
		struct MatchItem *match_item;

		yy_hconf = ptr->data;

		/* yy_conf == NULL is a fatal error for this connect block! */
		if ((yy_conf != NULL) && (yy_conf->name != NULL)) {
			new_hub_conf = make_conf_item(HUB_TYPE);
			match_item = (struct MatchItem *)map_to_conf(new_hub_conf);
			DupString(new_hub_conf->name, yy_conf->name);
			if (yy_hconf->user != NULL)
				DupString(match_item->user, yy_hconf->user);
			else
				DupString(match_item->user, "*");
			if (yy_hconf->host != NULL)
				DupString(match_item->host, yy_hconf->host);
			else
				DupString(match_item->host, "*");
		}
		dlinkDelete(&yy_hconf->node, &hub_conf_list);
		free_collect_item(yy_hconf);
	}

	/* Ditto for the LEAF confs */

	DLINK_FOREACH_SAFE(ptr, next_ptr, leaf_conf_list.head) {
		struct ConfItem *new_leaf_conf;
		struct MatchItem *match_item;

		yy_lconf = ptr->data;

		if ((yy_conf != NULL) && (yy_conf->name != NULL)) {
			new_leaf_conf = make_conf_item(LEAF_TYPE);
			match_item = (struct MatchItem *)map_to_conf(new_leaf_conf);
			DupString(new_leaf_conf->name, yy_conf->name);
			if (yy_lconf->user != NULL)
				DupString(match_item->user, yy_lconf->user);
			else
				DupString(match_item->user, "*");
			if (yy_lconf->host != NULL)
				DupString(match_item->host, yy_lconf->host);
			else
				DupString(match_item->host, "*");
		}
		dlinkDelete(&yy_lconf->node, &leaf_conf_list);
		free_collect_item(yy_lconf);
	}
	MyFree(class_name);
	class_name = NULL;
	yy_conf = NULL;
	yy_aconf = NULL;

	return 0;
}

static void
conf_set_connect_name(void *data)
{
	if (yy_conf->name != NULL) {
		conf_report_error("Warning -- connect::name specified multiple times.");
	}
	MyFree(yy_conf->name);
	DupString(yy_conf->name, data);
}

static void
conf_set_connect_host(void *data)
{
	MyFree(yy_aconf->host);
	DupString(yy_aconf->host, data);
}

static void
conf_set_connect_flags(void *data)
{
	conf_parm_t    *args = data;

	set_modes_from_table(&yy_aconf->flags, "flag", connect_table, args);
}

static void
conf_set_connect_send_password(void *data)
{
	if (yy_aconf->spasswd)
		memset(yy_aconf->spasswd, 0, strlen(yy_aconf->spasswd));
	MyFree(yy_aconf->spasswd);
	DupString(yy_aconf->spasswd, data);
}

static void
conf_set_connect_accept_password(void *data)
{
	if (yy_aconf->passwd)
		memset(yy_aconf->passwd, 0, strlen(yy_aconf->passwd));
	MyFree(yy_aconf->passwd);
	DupString(yy_aconf->passwd, data);
}

static void
conf_set_connect_port(void *data)
{
	yy_aconf->port = *(unsigned int *)data;
}

static void
conf_set_connect_aftype(void *data)
{
	char           *aft = data;

	if (strcasecmp(aft, "ipv4") == 0)
		yy_aconf->aftype = AF_INET;
#ifdef IPV6
	else if (strcasecmp(aft, "ipv6") == 0)
		yy_aconf->aftype = AF_INET6;
#endif
	else
		conf_report_error("connect::aftype '%s' is unknown.", aft);
}

static void
conf_set_connect_fakename(void *data)
{
	MyFree(yy_aconf->fakename);
	DupString(yy_aconf->fakename, data);
}

static void
conf_set_connect_encrypted(void *data)
{
	int		yesno = *(unsigned int *)data;

	if (yesno)
		yy_aconf->flags |= CONF_FLAGS_ENCRYPTED;
	else
		yy_aconf->flags &= ~CONF_FLAGS_ENCRYPTED;
}

static void
conf_set_connect_rsa_public_key_file(void *data)
{
#ifdef HAVE_LIBCRYPTO
	BIO            *file;

	if (yy_aconf->rsa_public_key) {
		RSA_free(yy_aconf->rsa_public_key);
		yy_aconf->rsa_public_key = NULL;
	}
	if (yy_aconf->rsa_public_key_file) {
		MyFree(yy_aconf->rsa_public_key_file);
		yy_aconf->rsa_public_key_file = NULL;
	}
	DupString(yy_aconf->rsa_public_key_file, data);

	file = BIO_new_file(data, "r");

	if (file == NULL) {
		conf_report_error
			("Ignoring connect::rsa_public_key_file %s -- couldn't open file.", data);
		return;
	}
	yy_aconf->rsa_public_key = (RSA *) PEM_read_bio_RSA_PUBKEY(file, NULL, 0, NULL);

	if (yy_aconf->rsa_public_key == NULL) {
		conf_report_error
			("Ignoring connect::rsa_public_key_file -- Key invalid; check syntax.");
		return;
	}
	BIO_set_close(file, BIO_CLOSE);
	BIO_free(file);
#else
	conf_report_error
	("Ignoring connect::rsa_public_key_file -- OpenSSL support not available.");
#endif
}

static void
conf_set_connect_cryptlink(void *data)
{
	int		yesno = *(unsigned int *)data;

	if (yesno)
		yy_aconf->flags |= CONF_FLAGS_CRYPTLINK;
	else
		yy_aconf->flags &= ~CONF_FLAGS_CRYPTLINK;
}

static void
conf_set_connect_compressed(void *data)
{
#ifdef HAVE_LIBZ
	int		yesno = *(unsigned int *)data;

	if (yesno)
		yy_aconf->flags |= CONF_FLAGS_COMPRESSED;
	else
		yy_aconf->flags &= ~CONF_FLAGS_COMPRESSED;
#else
			conf_report_error("Ignoring connect::compressed -- zlib not available.");
#endif
}

static void
conf_set_connect_auto(void *data)
{
	int		yesno = *(unsigned int *)data;

	if (yesno)
		yy_aconf->flags |= CONF_FLAGS_ALLOW_AUTO_CONN;
	else
		yy_aconf->flags &= ~CONF_FLAGS_ALLOW_AUTO_CONN;
}

static void
conf_set_connect_hub_mask(void *data)
{
	struct CollectItem *yy_tmp;

	yy_tmp = (struct CollectItem *)MyMalloc(sizeof(struct CollectItem));
	DupString(yy_tmp->host, (char *)data);
	DupString(yy_tmp->user, "*");
	dlinkAdd(yy_tmp, &yy_tmp->node, &hub_conf_list);
}

static void
conf_set_connect_leaf_mask(void *data)
{
	struct CollectItem *yy_tmp;

	yy_tmp = (struct CollectItem *)MyMalloc(sizeof(struct CollectItem));
	DupString(yy_tmp->host, (char *)data);
	DupString(yy_tmp->user, "*");
	dlinkAdd(yy_tmp, &yy_tmp->node, &leaf_conf_list);
}

static void
conf_set_connect_class(void *data)
{
	MyFree(class_name);
	DupString(class_name, (char *)data);
}

static void
conf_set_connect_cipher_preference(void *data)
{
#if 0
	struct EncCapability *ecap;
	char           *cipher_name;
	int		found = 0;

	yy_aconf->cipher_preference = NULL;

	cipher_name = data;

	for (ecap = CipherTable; ecap->name; ecap++) {
		if ((!irccmp(ecap->name, cipher_name)) && (ecap->cap & CAP_ENC_MASK)) {
			yy_aconf->cipher_preference = ecap;
			found = 1;
		}
	}

	if (!found) {
		conf_report_error("Invalid cipher '%s'.", cipher_name);
	}
#else
	conf_report_error("Ignoring connect::cipher_preference -- OpenSSL support not available.");
#endif
}

static int
conf_begin_kill(struct TopConf *tc)
{
	yy_conf = make_conf_item(KLINE_TYPE);
	yy_aconf = (struct AccessItem *)map_to_conf(yy_conf);

	return 0;
}

static int
conf_end_kill(struct TopConf *tc)
{
	if (yy_aconf->user && yy_aconf->reason && yy_aconf->host) {
		if (yy_aconf->host != NULL)
			add_conf_by_address(CONF_KILL, yy_aconf);
	} else
		delete_conf_item(yy_conf);

	yy_conf = NULL;
	yy_aconf = NULL;

	return 0;
}

static void
conf_set_kill_user(void *data)
{
	DupString(yy_aconf->host, data);
	split_user_host(yy_aconf->host, &yy_aconf->user, &yy_aconf->host);
}

static void
conf_set_kill_reason(void *data)
{
	MyFree(yy_aconf->reason);
	DupString(yy_aconf->reason, (char *)data);
}

static int
conf_begin_deny(struct TopConf *tc)
{
	yy_conf = make_conf_item(DLINE_TYPE);
	yy_aconf = (struct AccessItem *)map_to_conf(yy_conf);

	DupString(yy_aconf->reason, "No Reason");

	return 0;
}

static int
conf_end_deny(struct TopConf *tc)
{
	if (yy_aconf->host && parse_netmask(yy_aconf->host, NULL, NULL) != HM_HOST)
		add_conf_by_address(CONF_DLINE, yy_aconf);
	else
		delete_conf_item(yy_conf);

	yy_conf = NULL;
	yy_aconf = NULL;

	return 0;
}

static void
conf_set_deny_ip(void *data)
{
	MyFree(yy_aconf->host);
	DupString(yy_aconf->host, data);
}

static void
conf_set_deny_reason(void *data)
{
	MyFree(yy_aconf->reason);
	DupString(yy_aconf->reason, (char *)data);
}

static int
conf_begin_exempt(struct TopConf *tc)
{
	yy_conf = make_conf_item(EXEMPTDLINE_TYPE);
	yy_aconf = (struct AccessItem *)map_to_conf(yy_conf);
	DupString(yy_aconf->passwd, "*");

	return 0;
}

static int
conf_end_exempt(struct TopConf *tc)
{
	if (yy_aconf->host && parse_netmask(yy_aconf->host, NULL, NULL) != HM_HOST)
		add_conf_by_address(CONF_EXEMPTDLINE, yy_aconf);
	else
		delete_conf_item(yy_conf);
	yy_conf = NULL;
	yy_aconf = NULL;

	return 0;
}

static void
conf_set_exempt_ip(void *data)
{
	MyFree(yy_aconf->host);
	DupString(yy_aconf->host, (char *)data);
}

static int
conf_begin_gecos(struct TopConf *tc)
{
	yy_conf = make_conf_item(XLINE_TYPE);
	yy_match_item = map_to_conf(yy_conf);
	DupString(yy_match_item->reason, "Something about your name");

	return 0;
}

static void
conf_set_gecos_name(void *data)
{
	MyFree(yy_conf->name);
	DupString(yy_conf->name, data);
	collapse(yy_conf->name);
}

static void
conf_set_gecos_reason(void *data)
{
	MyFree(yy_match_item->reason);
	DupString(yy_match_item->reason, data);
}

static void
conf_set_gecos_action(void *data)
{
	char           *act = data;

	if (strcasecmp(act, "warn") == 0)
		yy_match_item->action = 0;
	else if (strcasecmp(act, "reject") == 0)
		yy_match_item->action = 1;
	else if (strcasecmp(act, "silent") == 0)
		yy_match_item->action = 2;
	else
		conf_report_error("Warning -- invalid gecos::action.");
}

/* begin alias stuff */
static int
conf_begin_alias(struct TopConf *tc)
{
	yy_conf = make_conf_item(ALIAS_TYPE);
	yy_alconf = (struct Alias *)map_to_conf(yy_conf);

	return 0;
}

static int
conf_end_alias(struct TopConf *tc)
{
	struct ConfItem *yy_tmp;
	dlink_node     *ptr, *next_ptr;

	if (conf_cur_block_name != NULL) {
		MyFree(yy_conf->name);
		DupString(yy_conf->name, conf_cur_block_name);
	}
	if (EmptyString(yy_conf->name)) {
		conf_report_error("Ignoring alias block -- missing name");
		return 0;
	}
	/*
	 * Now, make sure there is a copy of the "base" given oper block in
	 * each of the collected copies
	 */

	DLINK_FOREACH_SAFE(ptr, next_ptr, alconf_items.head) {
		struct Alias   *new_alconf;
		struct Alias   *zz_tmp;
		struct ConfItem *new_conf;
		yy_tmp = ptr->data;

		new_conf = make_conf_item(ALIAS_TYPE);
		new_alconf = (struct Alias *)map_to_conf(new_conf);
		zz_tmp = (struct Alias *)map_to_conf(yy_tmp);

		if (new_alconf->name != NULL)
			DupString(new_alconf->name, zz_tmp->name);
		if (new_alconf->command != NULL)
			DupString(new_alconf->command, zz_tmp->command);

		new_alconf->override = zz_tmp->override;
		dlinkDelete(&yy_tmp->node, &alconf_items);
		free_alias_item((struct Alias *)yy_tmp);
	}

	yy_conf = NULL;
	yy_alconf = NULL;

	return 0;
}

static void
conf_set_alias_command(void *data)
{

	/* Check the line for IRC compliancy */
	if ((strlen((const char *)data)) > 512)
		((char *)data)[512] = '\0';

	if (yy_alconf)
		DupString(yy_alconf->command, (char *)data);
}

static void
conf_set_alias_override(void *data)
{
	if (yy_alconf)
		yy_alconf->override = (int)data;
}

/* end alias stuff */

static void
conf_set_general_failed_oper_notice(void *data)
{
	ConfigFileEntry.failed_oper_notice = *(unsigned int *)data;
}

static void
conf_set_general_throttle_time(void *data)
{
	ConfigFileEntry.throttle_time = *(unsigned int *)data;
}

static void
conf_set_general_max_silence(void *data)
{
	ConfigFileEntry.max_silence = *(unsigned int *)data;
}

static void
conf_set_general_maximum_links(void *data)
{
	ConfigFileEntry.maximum_links = *(unsigned int *)data;
}

static void
conf_set_general_anti_nick_flood(void *data)
{
	ConfigFileEntry.anti_nick_flood = *(unsigned int *)data;
}

static void
conf_set_general_max_nick_time(void *data)
{
	ConfigFileEntry.max_nick_time = *(unsigned int *)data;
}

static void
conf_set_general_max_nick_changes(void *data)
{
	ConfigFileEntry.max_nick_changes = *(unsigned int *)data;
}

static void
conf_set_general_max_accept(void *data)
{
	ConfigFileEntry.max_accept = *(unsigned int *)data;
}

static void
conf_set_general_anti_spam_exit_message_time(void *data)
{
	ConfigFileEntry.anti_spam_exit_message_time = *(unsigned int *)data;
}

static void
conf_set_general_ts_warn_delta(void *data)
{
	ConfigFileEntry.ts_warn_delta = *(unsigned int *)data;
}

static void
conf_set_general_ts_max_delta(void *data)
{
	ConfigFileEntry.ts_max_delta = *(unsigned int *)data;
}

static void
conf_set_general_havent_read_conf(void *data)
{
	if (*(unsigned int *)data) {
		conf_report_error("You haven't read your config file properly.");
		conf_report_error
			("There is a line in the example conf that will kill your server if not removed.");
		conf_report_error
			("Consider actually reading/editing the conf file, and removing this line.");
		if (!testing_conf)
			exit(0);
	}
}

static void
conf_set_general_kline_with_reason(void *data)
{
	ConfigFileEntry.kline_with_reason = *(unsigned int *)data;
}

static void
conf_set_general_permanant_simbans(void *data)
{
	ConfigFileEntry.permanant_simbans = *(unsigned int *)data;
}

static void
conf_set_general_allow_old_protocol(void *data)
{
        ConfigFileEntry.allow_old_protocol = *(unsigned int *)data;
}

static void
conf_set_hiding_hide_netsplit_source(void *data)
{
        ConfigFileEntry.hide_netsplit_source = *(unsigned int *)data;
}

static void
conf_set_hiding_hidden_server(void *data)
{
        ConfigFileEntry.hidden_server = *(unsigned int *)data;
}

static void
conf_set_hiding_disable_hiding(void *data)
{
        ConfigFileEntry.disable_hiding = *(unsigned int *)data;
}

static void
conf_set_general_verbose(void *data)
{
	ConfigFileEntry.verbose = *(unsigned int *)data;
}

static void
conf_set_general_kline_with_connection_closed(void *data)
{
	ConfigFileEntry.kline_with_connection_closed = *(unsigned int *)data;
}

static void
conf_set_general_warn_no_nline(void *data)
{
	ConfigFileEntry.warn_no_nline = *(unsigned int *)data;
}

static void
conf_set_general_stats_o_oper_only(void *data)
{
	ConfigFileEntry.stats_o_oper_only = *(unsigned int *)data;
}

static void
conf_set_general_stats_P_oper_only(void *data)
{
	ConfigFileEntry.stats_P_oper_only = *(unsigned int *)data;
}

static void
conf_set_general_stats_k_oper_only(void *data)
{
	char           *val = data;

	if (strcasecmp(val, "yes") == 0)
		ConfigFileEntry.stats_k_oper_only = 2;
	else if (strcasecmp(val, "masked") == 0)
		ConfigFileEntry.stats_k_oper_only = 1;
	else if (strcasecmp(val, "no") == 0)
		ConfigFileEntry.stats_k_oper_only = 0;
	else
		conf_report_error("Invalid setting '%s' for general::stats_k_oper_only.", val);
}

static void
conf_set_general_stats_i_oper_only(void *data)
{
	char           *val = data;

	if (strcasecmp(val, "yes") == 0)
		ConfigFileEntry.stats_i_oper_only = 2;
	else if (strcasecmp(val, "masked") == 0)
		ConfigFileEntry.stats_i_oper_only = 1;
	else if (strcasecmp(val, "no") == 0)
		ConfigFileEntry.stats_i_oper_only = 0;
	else
		conf_report_error("Invalid setting '%s' for general::stats_i_oper_only.", val);
}

static void
conf_set_general_pace_wait(void *data)
{
	ConfigFileEntry.pace_wait = *(unsigned int *)data;
}

static void
conf_set_general_caller_id_wait(void *data)
{
	ConfigFileEntry.caller_id_wait = *(unsigned int *)data;
}

static void
conf_set_general_pace_wait_simple(void *data)
{
	ConfigFileEntry.pace_wait_simple = *(unsigned int *)data;
}

static void
conf_set_general_short_motd(void *data)
{
	ConfigFileEntry.short_motd = *(unsigned int *)data;
}

static void
conf_set_general_no_oper_flood(void *data)
{
	ConfigFileEntry.no_oper_flood = *(unsigned int *)data;
}

static void
conf_set_general_message_locale(void *data)
{
	char		langenv   [BUFSIZE];

	if (strlen((char *)data) > BUFSIZE - 10)
		((char *)data)[BUFSIZE - 9] = 0;

	ircsprintf(langenv, "LANGUAGE=%s", (char *)data);
	putenv(langenv);
}

static void
conf_set_general_idletime(void *data)
{
	ConfigFileEntry.idletime = *(unsigned int *)data;
}

static void
conf_set_general_dots_in_ident(void *data)
{
	ConfigFileEntry.dots_in_ident = *(unsigned int *)data;
}

static void
conf_set_general_max_targets(void *data)
{
	ConfigFileEntry.max_targets = *(unsigned int *)data;
}

static void
conf_set_general_servlink_path(void *data)
{
	MyFree(ConfigFileEntry.servlink_path);
	DupString(ConfigFileEntry.servlink_path, data);
}

static void
conf_set_general_default_cipher_preference(void *data)
{
#if 0
	struct EncCapability *ecap;
	char           *cipher_name;

	ConfigFileEntry.default_cipher_preference = NULL;

	cipher_name = data;

	for (ecap = CipherTable; ecap->name; ecap++) {
		if ((!irccmp(ecap->name, cipher_name)) && (ecap->cap & CAP_ENC_MASK)) {
			ConfigFileEntry.default_cipher_preference = ecap;
			return;
		}
	}

	conf_report_error("Invalid general::default_cipher_preference '%s'.", cipher_name);
#else
	conf_report_error
	("Ignoring general::default_cipher_preference -- OpenSSL support not available.");
#endif
}

static void
conf_set_general_compression_level(void *data)
{
#ifdef HAVE_LIBZ
	ConfigFileEntry.compression_level = *(unsigned int *)data;

	if ((ConfigFileEntry.compression_level < 1) || (ConfigFileEntry.compression_level > 9)) {
		conf_report_error
			("Invalid general::compression_level %d -- using default.",
			 ConfigFileEntry.compression_level);
		ConfigFileEntry.compression_level = 0;
	}
#else
	conf_report_error("Ignoring general::compression_level -- zlib not available.");
#endif
}

static void
conf_set_general_use_egd(void *data)
{
	ConfigFileEntry.use_egd = *(unsigned int *)data;
}

static void
conf_set_general_egdpool_path(void *data)
{
	MyFree(ConfigFileEntry.egdpool_path);
	DupString(ConfigFileEntry.egdpool_path, data);
}

static void
conf_set_general_ping_cookie(void *data)
{
	ConfigFileEntry.ping_cookie = *(unsigned int *)data;
}

static void
conf_set_general_disable_auth(void *data)
{
	ConfigFileEntry.disable_auth = *(unsigned int *)data;
}

static void
conf_set_general_min_nonwildcard(void *data)
{
	ConfigFileEntry.min_nonwildcard = *(unsigned int *)data;
}

static void
conf_set_general_min_nonwildcard_simple(void *data)
{
	ConfigFileEntry.min_nonwildcard_simple = *(unsigned int *)data;
}

static void
conf_set_general_default_floodcount(void *data)
{
	ConfigFileEntry.default_floodcount = *(unsigned int *)data;
}

static void
conf_set_general_oper_pass_resv(void *data)
{
	int		yesno = *(int *)data;
	ConfigFileEntry.oper_pass_resv = yesno;
}

static void
conf_set_general_true_no_oper_flood(void *data)
{
	int		yesno = *(int *)data;
	ConfigFileEntry.true_no_oper_flood = yesno;
}

static void
conf_set_general_anti_cgi_irc(void *data)
{
	int		yesno = *(int *)data;
	ConfigFileEntry.anti_cgi_irc = yesno;
}

static void
conf_set_general_disable_dns(void *data)
{
	int		yesno = *(int *)data;
	ConfigFileEntry.disable_dns = yesno;
}

static void
conf_set_general_disable_remote_commands(void *data)
{
	int		yesno = *(int *)data;
	ConfigFileEntry.disable_remote = yesno;
}

static void
conf_set_general_client_flood(void *data)
{
	ConfigFileEntry.client_flood = *(unsigned int *)data;
}

static void
conf_set_general_dot_in_ip6_addr(void *data)
{
	ConfigFileEntry.dot_in_ip6_addr = *(unsigned int *)data;
}

static void
conf_set_channel_use_except(void *data)
{
	ConfigChannel.use_except = *(unsigned int *)data;
}

static void
conf_set_channel_use_invex(void *data)
{
	ConfigChannel.use_invex = *(unsigned int *)data;
}

static void
conf_set_channel_use_knock(void *data)
{
	ConfigChannel.use_knock = *(unsigned int *)data;
}

static void
conf_set_channel_knock_delay(void *data)
{
	ConfigChannel.knock_delay = *(unsigned int *)data;
}

static void
conf_set_channel_knock_delay_channel(void *data)
{
	ConfigChannel.knock_delay_channel = *(unsigned int *)data;
}

static void
conf_set_channel_max_chans_per_user(void *data)
{
	ConfigChannel.max_chans_per_user = *(unsigned int *)data;
}

static void
conf_set_channel_quiet_on_ban(void *data)
{
	ConfigChannel.quiet_on_ban = *(unsigned int *)data;
}

static void
conf_set_channel_max_bans(void *data)
{
	ConfigChannel.max_bans = *(unsigned int *)data;
}


static void
conf_set_channel_default_split_user_count(void *data)
{
	ConfigChannel.default_split_user_count = *(unsigned int *)data;
}

static void
conf_set_channel_default_split_server_count(void *data)
{
	ConfigChannel.default_split_server_count = *(unsigned int *)data;
}

static void
conf_set_channel_no_create_on_split(void *data)
{
	ConfigChannel.no_create_on_split = *(unsigned int *)data;
}

static void
conf_set_channel_no_join_on_split(void *data)
{
	ConfigChannel.no_join_on_split = *(unsigned int *)data;
}

static int
conf_end_network(struct TopConf *tc)
{
	if (conf_cur_block_name != NULL) {
		char           *p;

		if ((p = strchr(conf_cur_block_name, ' ')) != NULL)
			p = '\0';

		MyFree(ServerInfo.network_name);
		DupString(ServerInfo.network_name, conf_cur_block_name);
	}
	return 0;
}

static void
conf_set_network_name(void *data)
{
	if (data != NULL) {
		char           *p;

		if ((p = strchr((char *)data, ' ')) != NULL)
			p = '\0';

		MyFree(ServerInfo.network_name);
		DupString(ServerInfo.network_name, (char *)data);
	}
}

static void
conf_set_network_description(void *data)
{
	MyFree(ServerInfo.network_desc);
	DupString(ServerInfo.network_desc, (char *)data);
}

static void
conf_set_network_on_oper_host(void *data)
{
	char           *p;

	if ((p = strchr((char *)data, ' ')) != NULL)
		p = '\0';

	MyFree(ServerInfo.network_operhost);
	DupString(ServerInfo.network_operhost, (char *)data);
}

static void
conf_set_network_gline_address(void *data)
{
	char           *p;

	if ((p = strchr((char *)data, ' ')) != NULL)
		p = '\0';

	MyFree(ServerInfo.network_glineaddr);
	DupString(ServerInfo.network_glineaddr, (char *)data);
}

static void
conf_set_network_cloak_key_1(void *data)
{
	MyFree(ServerInfo.network_cloakkey1);
	DupString(ServerInfo.network_cloakkey1, (char *)data);
}

static void
conf_set_network_cloak_key_2(void *data)
{
	MyFree(ServerInfo.network_cloakkey2);
	DupString(ServerInfo.network_cloakkey2, (char *)data);
}

static void
conf_set_network_cloak_key_3(void *data)
{
	MyFree(ServerInfo.network_cloakkey3);
	DupString(ServerInfo.network_cloakkey3, (char *)data);
}

static void
conf_set_network_cloak_on_oper(void *data)
{
	int		yesno = *(int *)data;
	ServerInfo.network_cloak_on_oper = yesno;
}

static void
conf_set_network_oper_prefix(void *data)
{
	int		yesno = *(int *)data;
	ConfigFileEntry.oper_prefix = yesno;
}

static void
conf_set_network_disable_hostmasking(void *data)
{
	int		yesno = *(int *)data;
	ConfigFileEntry.disable_hostmasking = yesno;
}

static void
conf_set_network_tidy_quits(void *data)
{
	int		yesno = *(int *)data;
	ConfigFileEntry.tidy_quits = yesno;
}

static void
conf_set_network_cloak_on_connect(void *data)
{
	int		yesno = *(int *)data;
	ServerInfo.network_cloak_on_connect = yesno;
}

static void
conf_set_network_filter_on_connect(void *data)
{
	int		yesno = *(int *)data;
	ConfigFileEntry.filter_on_connect = yesno;
}

static void
conf_set_wingate_enable(void *data)
{
	int		yesno = *(int *)data;
	ServerInfo.wingate_enable = yesno;
}

static void
conf_set_wingate_monitorbot(void *data)
{
	MyFree(ServerInfo.monitorbot);
	DupString(ServerInfo.monitorbot, (char *)data);
}

static void
conf_set_wingate_webpage(void *data)
{
	MyFree(ServerInfo.wingate_website);
	DupString(ServerInfo.wingate_website, (char *)data);
}

static void
conf_set_hiding_whois_hiding(void *data)
{
	MyFree(ServerInfo.whois_hiding);
	DupString(ServerInfo.whois_hiding, (char *)data);
}

static void
conf_set_hiding_whois_description(void *data)
{
	MyFree(ServerInfo.whois_description);
	DupString(ServerInfo.whois_description, (char *)data);
}

static void
conf_set_hiding_roundrobin(void *data)
{
	MyFree(ServerInfo.roundrobin);
	DupString(ServerInfo.roundrobin, (char *)data);
}

static void
conf_set_badwords_filter(void *data)
{
	filter_del_word((char *)data);
	filter_add_word((char *)data);
}

void
conf_report_error(const char *fmt,...)
{
	va_list		ap;
	char		msg       [IRCD_BUFSIZE + 1] = {0};
	extern char    *current_file;

	va_start(ap, fmt);
	vsnprintf(msg, IRCD_BUFSIZE, fmt, ap);
	va_end(ap);

	if (testing_conf) {
		fprintf(stderr, "\"%s\", line %d: %s\n", current_file, lineno + 1, msg);
		return;
	}
	ilog(L_ERROR, "\"%s\", line %d: %s", current_file, lineno + 1, msg);
	sendto_realops_flags(UMODE_ALL, L_ALL, "\"%s\", line %d: %s", current_file,
			     lineno + 1, msg);
}

int
conf_start_block(const char *block, const char *name)
{
	if ((conf_cur_block = find_top_conf(block)) == NULL) {
		conf_report_error("Configuration block '%s' is not defined.", block);
		return -1;
	}
	if (name)
		DupString(conf_cur_block_name, name);
	else
		conf_cur_block_name = NULL;

	if (conf_cur_block->tc_sfunc)
		if (conf_cur_block->tc_sfunc(conf_cur_block) < 0)
			return -1;

	return 0;
}

int
conf_end_block(struct TopConf *tc)
{
	if (tc->tc_efunc)
		return tc->tc_efunc(tc);

	MyFree(conf_cur_block_name);
	return 0;
}

int
conf_call_set(struct TopConf *tc, const char *item, conf_parm_t * value, int type)
{
	struct ConfEntry *cf;
	conf_parm_t    *cp;

	if (!tc)
		return -1;

	if ((cf = find_conf_item(tc, item)) == NULL) {
		conf_report_error("Non-existant configuration setting %s::%s.",
				  tc->tc_name, (char *)item);
		return -1;
	}
	/*
	 * if it takes one thing, make sure they only passed one thing, and
	 * handle as needed.
	 */
	if (value->type & CF_FLIST && !cf->cf_type & CF_FLIST) {
		conf_report_error("Option %s::%s does not take a list of values.",
				  tc->tc_name, item);
		return -1;
	}
	cp = value->v.list;


	if (CF_TYPE(value->v.list->type) != CF_TYPE(cf->cf_type)) {
		/*
		 * if it expects a string value, but we got a yesno, convert
		 * it back
		 */
		if ((CF_TYPE(value->v.list->type) == CF_YESNO) &&
		    (CF_TYPE(cf->cf_type) == CF_STRING)) {
			value->v.list->type = CF_STRING;

			if (cp->v.number == 1)
				DupString(cp->v.string, "yes");
			else
				DupString(cp->v.string, "no");
		}
		/*
		 * maybe it's a CF_TIME and they passed CF_INT -- should
		 * still be valid
		 */
		else if (!((CF_TYPE(value->v.list->type) == CF_INT) &&
			   (CF_TYPE(cf->cf_type) == CF_TIME))) {
			conf_report_error("Wrong type for %s::%s (expected %s, got %s)",
					  tc->tc_name, (char *)item,
					  conf_strtype(cf->cf_type),
					  conf_strtype(value->v.list->type));
			return -1;
		}
	}
	if (cf->cf_type & CF_FLIST) {
		/* just pass it the extended argument list */
		cf->cf_func(value->v.list);
	} else {
		/* it's old-style, needs only one arg */
		switch (cf->cf_type) {
		case CF_INT:
		case CF_TIME:
		case CF_YESNO:
			cf->cf_func(&cp->v.number);
			break;
		case CF_STRING:
		case CF_QSTRING:
			cf->cf_func(cp->v.string);
			break;
		}
	}


	return 0;
}

int
add_conf_item(const char *topconf, const char *name, int type, void (*func) (void *))
{
	struct TopConf *tc;
	struct ConfEntry *cf;

	dlink_node     *ptr = make_dlink_node();

	if ((tc = find_top_conf(topconf)) == NULL)
		return -1;

	if ((cf = find_conf_item(tc, name)) != NULL)
		return -1;

	cf = MyMalloc(sizeof(struct ConfEntry));

	DupString(cf->cf_name, name);
	cf->cf_type = type;
	cf->cf_func = func;

	dlinkAdd(cf, ptr, &tc->tc_items);

	return 0;
}

int
remove_conf_item(const char *topconf, const char *name)
{
	struct TopConf *tc;
	struct ConfEntry *cf;
	dlink_node     *ptr;

	if ((tc = find_top_conf(topconf)) == NULL)
		return -1;

	if ((cf = find_conf_item(tc, name)) == NULL)
		return -1;

	if ((ptr = dlinkFind(&tc->tc_items, cf)) == NULL)
		return -1;

	dlinkDelete(ptr, &tc->tc_items);
	MyFree(cf->cf_name);
	MyFree(cf);

	return 0;
}

void
newconf_init()
{
	add_top_conf("modules", NULL, NULL);
	add_conf_item("modules", "path", CF_QSTRING, conf_set_modules_path);
	add_conf_item("modules", "module", CF_QSTRING, conf_set_modules_module);

	add_top_conf("serverinfo", NULL, NULL);
	add_conf_item("serverinfo", "rsa_private_key_file", CF_QSTRING,
		      conf_set_serverinfo_rsa_private_key_file);
	add_conf_item("serverinfo", "ssl_certificate_file", CF_QSTRING,
		      conf_set_serverinfo_ssl_certificate_file);
	add_conf_item("serverinfo", "ssl_ca_certificate_file", CF_QSTRING,
		      conf_set_serverinfo_ssl_ca_certificate_file);
	add_conf_item("serverinfo", "name", CF_QSTRING, conf_set_serverinfo_name);
	add_conf_item("serverinfo", "description", CF_QSTRING, conf_set_serverinfo_description);
	add_conf_item("serverinfo", "vhost", CF_QSTRING, conf_set_serverinfo_vhost);
	add_conf_item("serverinfo", "vhost6", CF_QSTRING, conf_set_serverinfo_vhost6);
	add_conf_item("serverinfo", "hub", CF_YESNO, conf_set_serverinfo_hub);
	add_conf_item("serverinfo", "sid", CF_QSTRING, conf_set_serverinfo_sid);
	add_conf_item("serverinfo", "max_clients", CF_INT, conf_set_serverinfo_max_clients);

	add_top_conf("network", NULL, conf_end_network);
	add_conf_item("network", "name", CF_QSTRING, conf_set_network_name);
	add_conf_item("network", "description", CF_QSTRING, conf_set_network_description);
	add_conf_item("network", "on_oper_host", CF_QSTRING, conf_set_network_on_oper_host);
	add_conf_item("network", "gline_address", CF_QSTRING, conf_set_network_gline_address);
	add_conf_item("network", "cloak_key_1", CF_QSTRING, conf_set_network_cloak_key_1);
	add_conf_item("network", "cloak_key_2", CF_QSTRING, conf_set_network_cloak_key_2);
	add_conf_item("network", "cloak_key_3", CF_QSTRING, conf_set_network_cloak_key_3);
	add_conf_item("network", "cloak_on_oper", CF_YESNO, conf_set_network_cloak_on_oper);
	add_conf_item("network", "oper_prefix", CF_YESNO, conf_set_network_oper_prefix);
	add_conf_item("network", "disable_hostmasking", CF_YESNO,
		      conf_set_network_disable_hostmasking);
	add_conf_item("network", "clean_quits", CF_YESNO, conf_set_network_tidy_quits);
	add_conf_item("network", "cloak_on_connect", CF_YESNO, conf_set_network_cloak_on_connect);
	add_conf_item("network", "filter_on_connect", CF_YESNO, conf_set_network_filter_on_connect);

	add_top_conf("admin", NULL, NULL);
	add_conf_item("admin", "name", CF_QSTRING, conf_set_admin_name);
	add_conf_item("admin", "description", CF_QSTRING, conf_set_admin_description);
	add_conf_item("admin", "email", CF_QSTRING, conf_set_admin_email);

	add_top_conf("operator", conf_begin_oper, conf_end_oper);
	add_conf_item("operator", "name", CF_QSTRING, conf_set_oper_name);
	add_conf_item("operator", "user", CF_QSTRING, conf_set_oper_user);
	add_conf_item("operator", "password", CF_QSTRING, conf_set_oper_password);
	add_conf_item("operator", "class", CF_QSTRING, conf_set_oper_class);
	add_conf_item("operator", "encrypted", CF_YESNO, conf_set_oper_encrypted);
	add_conf_item("operator", "rsa_public_key_file", CF_QSTRING,
		      conf_set_oper_rsa_public_key_file);
	add_conf_item("operator", "flags", CF_STRING | CF_FLIST, conf_set_oper_flags);
	add_conf_item("operator", "user_modes", CF_QSTRING, conf_set_oper_umodes);

	add_top_conf("class", conf_begin_class, conf_end_class);
	add_conf_item("class", "name", CF_QSTRING, conf_set_class_name);
	add_conf_item("class", "ping_time", CF_TIME, conf_set_class_ping_time);
	add_conf_item("class", "number_per_ip", CF_INT, conf_set_class_number_per_ip);
	add_conf_item("class", "number_per_ip_global", CF_INT, conf_set_class_number_per_ip_global);
	add_conf_item("class", "number_per_ident", CF_INT, conf_set_class_number_per_ident);
	add_conf_item("class", "connectfreq", CF_TIME, conf_set_class_connectfreq);
	add_conf_item("class", "max_number", CF_INT, conf_set_class_max_number);
	add_conf_item("class", "sendq", CF_TIME, conf_set_class_sendq);

	add_top_conf("listen", conf_begin_listen, conf_end_listen);
	add_conf_item("listen", "port", CF_INT | CF_FLIST, conf_set_listen_port);
	add_conf_item("listen", "sslport", CF_INT | CF_FLIST, conf_set_listen_sslport);
	add_conf_item("listen", "ip", CF_QSTRING, conf_set_listen_address);
	add_conf_item("listen", "host", CF_QSTRING, conf_set_listen_address);

	add_top_conf("auth", conf_begin_auth, conf_end_auth);
	add_conf_item("auth", "user", CF_QSTRING, conf_set_auth_user);
	add_conf_item("auth", "password", CF_QSTRING, conf_set_auth_passwd);
	add_conf_item("auth", "encrypted", CF_YESNO, conf_set_auth_encrypted);
	add_conf_item("auth", "class", CF_QSTRING, conf_set_auth_class);
	add_conf_item("auth", "kline_exempt", CF_YESNO, conf_set_auth_kline_exempt);
	add_conf_item("auth", "need_ident", CF_YESNO, conf_set_auth_need_ident);
	add_conf_item("auth", "have_ident", CF_YESNO, conf_set_auth_need_ident);
	add_conf_item("auth", "restricted", CF_YESNO, conf_set_auth_is_restricted);
	add_conf_item("auth", "exceed_limit", CF_YESNO, conf_set_auth_exceed_limit);
	add_conf_item("auth", "no_tilde", CF_YESNO, conf_set_auth_no_tilde);
	add_conf_item("auth", "spoof", CF_QSTRING, conf_set_auth_spoof);
	add_conf_item("auth", "spoof_notice", CF_YESNO, conf_set_auth_spoof_notice);
	add_conf_item("auth", "redirserv", CF_QSTRING, conf_set_auth_redir_serv);
	add_conf_item("auth", "redirport", CF_INT, conf_set_auth_redir_port);
	add_conf_item("auth", "flags", CF_STRING | CF_FLIST, conf_set_auth_flags);

	add_top_conf("resv", conf_begin_resv, conf_end_resv);
	add_conf_item("resv", "reason", CF_QSTRING, conf_set_resv_reason);
	add_conf_item("resv", "channel", CF_QSTRING, conf_set_resv_channel);
	add_conf_item("resv", "nick", CF_QSTRING, conf_set_resv_nick);

	add_top_conf("connect", conf_begin_connect, conf_end_connect);
	add_conf_item("connect", "name", CF_QSTRING, conf_set_connect_name);
	add_conf_item("connect", "host", CF_QSTRING, conf_set_connect_host);
	add_conf_item("connect", "send_password", CF_QSTRING, conf_set_connect_send_password);
	add_conf_item("connect", "accept_password", CF_QSTRING, conf_set_connect_accept_password);
	add_conf_item("connect", "port", CF_INT, conf_set_connect_port);
	add_conf_item("connect", "aftype", CF_STRING, conf_set_connect_aftype);
	add_conf_item("connect", "fakename", CF_QSTRING, conf_set_connect_fakename);
	add_conf_item("connect", "hub_mask", CF_QSTRING, conf_set_connect_hub_mask);
	add_conf_item("connect", "leaf_mask", CF_QSTRING, conf_set_connect_leaf_mask);
	add_conf_item("connect", "class", CF_QSTRING, conf_set_connect_class);
	add_conf_item("connect", "autoconn", CF_YESNO, conf_set_connect_auto);
	add_conf_item("connect", "encrypted", CF_YESNO, conf_set_connect_encrypted);
	add_conf_item("connect", "compressed", CF_YESNO, conf_set_connect_compressed);
	add_conf_item("connect", "cryptlink", CF_YESNO, conf_set_connect_cryptlink);
	add_conf_item("connect", "rsa_public_key_file", CF_QSTRING,
		      conf_set_connect_rsa_public_key_file);
	add_conf_item("connect", "cipher_preference", CF_QSTRING,
		      conf_set_connect_cipher_preference);
	add_conf_item("connect", "flags", CF_STRING | CF_FLIST, conf_set_connect_flags);

	add_top_conf("kill", conf_begin_kill, conf_end_kill);
	add_conf_item("kill", "user", CF_QSTRING, conf_set_kill_user);
	add_conf_item("kill", "reason", CF_QSTRING, conf_set_kill_reason);

	add_top_conf("deny", conf_begin_deny, conf_end_deny);
	add_conf_item("deny", "ip", CF_QSTRING, conf_set_deny_ip);
	add_conf_item("deny", "reason", CF_QSTRING, conf_set_deny_reason);

	add_top_conf("exempt", conf_begin_exempt, conf_end_exempt);
	add_conf_item("exempt", "ip", CF_QSTRING, conf_set_exempt_ip);

	add_top_conf("gecos", conf_begin_gecos, NULL);
	add_conf_item("gecos", "name", CF_QSTRING, conf_set_gecos_name);
	add_conf_item("gecos", "reason", CF_QSTRING, conf_set_gecos_reason);
	add_conf_item("gecos", "action", CF_STRING, conf_set_gecos_action);

	add_top_conf("general", NULL, NULL);
	add_conf_item("general", "failed_oper_notice", CF_YESNO,
		      conf_set_general_failed_oper_notice);
	add_conf_item("general", "anti_nick_flood", CF_YESNO, conf_set_general_anti_nick_flood);
	add_conf_item("general", "max_nick_time", CF_TIME, conf_set_general_max_nick_time);
	add_conf_item("general", "max_nick_changes", CF_INT, conf_set_general_max_nick_changes);
	add_conf_item("general", "max_accept", CF_INT, conf_set_general_max_accept);
	add_conf_item("general", "anti_spam_exit_message_time", CF_TIME,
		      conf_set_general_anti_spam_exit_message_time);
	add_conf_item("general", "ts_warn_delta", CF_TIME, conf_set_general_ts_warn_delta);
	add_conf_item("general", "ts_max_delta", CF_TIME, conf_set_general_ts_max_delta);
	add_conf_item("general", "kline_with_reason", CF_YESNO, conf_set_general_kline_with_reason);
	add_conf_item("general", "kline_with_connection_closed", CF_YESNO,
		      conf_set_general_kline_with_connection_closed);
	add_conf_item("general", "warn_no_nline", CF_YESNO, conf_set_general_warn_no_nline);
	add_conf_item("general", "dots_in_ident", CF_INT, conf_set_general_dots_in_ident);
	add_conf_item("general", "stats_o_oper_only", CF_YESNO, conf_set_general_stats_o_oper_only);
	add_conf_item("general", "stats_k_oper_only", CF_STRING,
		      conf_set_general_stats_k_oper_only);
	add_conf_item("general", "stats_i_oper_only", CF_STRING,
		      conf_set_general_stats_i_oper_only);
	add_conf_item("general", "stats_P_oper_only", CF_YESNO, conf_set_general_stats_P_oper_only);
	add_conf_item("general", "pace_wait", CF_TIME, conf_set_general_pace_wait);
	add_conf_item("general", "pace_wait_simple", CF_TIME, conf_set_general_pace_wait_simple);
	add_conf_item("general", "short_motd", CF_YESNO, conf_set_general_short_motd);
	add_conf_item("general", "no_oper_flood", CF_YESNO, conf_set_general_no_oper_flood);
	add_conf_item("general", "idletime", CF_TIME, conf_set_general_idletime);
	add_conf_item("general", "message_locale", CF_QSTRING, conf_set_general_message_locale);
	add_conf_item("general", "max_targets", CF_INT, conf_set_general_max_targets);
	add_conf_item("general", "use_egd", CF_YESNO, conf_set_general_use_egd);
	add_conf_item("general", "egdpool_path", CF_QSTRING, conf_set_general_egdpool_path);
	add_conf_item("general", "caller_id_wait", CF_TIME, conf_set_general_caller_id_wait);
	add_conf_item("general", "default_floodcount", CF_INT, conf_set_general_default_floodcount);
	add_conf_item("general", "min_nonwildcard", CF_INT, conf_set_general_min_nonwildcard);
	add_conf_item("general", "min_nonwildcard_simple", CF_INT,
		      conf_set_general_min_nonwildcard_simple);
	add_conf_item("general", "servlink_path", CF_QSTRING, conf_set_general_servlink_path);
	add_conf_item("general", "default_cipher_preference", CF_QSTRING,
		      conf_set_general_default_cipher_preference);
	add_conf_item("general", "compression_level", CF_INT, conf_set_general_compression_level);
	add_conf_item("general", "client_flood", CF_TIME, conf_set_general_client_flood);
	add_conf_item("general", "havent_read_conf", CF_YESNO, conf_set_general_havent_read_conf);
	add_conf_item("general", "dot_in_ip6_addr", CF_YESNO, conf_set_general_dot_in_ip6_addr);
	add_conf_item("general", "ping_cookie", CF_YESNO, conf_set_general_ping_cookie);
	add_conf_item("general", "disable_auth", CF_YESNO, conf_set_general_disable_auth);
	add_conf_item("general", "throttle_time", CF_INT, conf_set_general_throttle_time);
	add_conf_item("general", "max_silence", CF_INT, conf_set_general_max_silence);
	add_conf_item("general", "maximum_links", CF_INT, conf_set_general_maximum_links);
	add_conf_item("general", "oper_pass_resv", CF_YESNO, conf_set_general_oper_pass_resv);
	add_conf_item("general", "true_no_oper_flood", CF_YESNO,
		      conf_set_general_true_no_oper_flood);
	add_conf_item("general", "anti_cgi_irc", CF_YESNO, conf_set_general_anti_cgi_irc);
	add_conf_item("general", "disable_dns", CF_YESNO, conf_set_general_disable_dns);
	add_conf_item("general", "disable_remote_commands", CF_YESNO,
		      conf_set_general_disable_remote_commands);
	add_conf_item("general", "permanant_simbans", CF_YESNO, conf_set_general_permanant_simbans);
	add_conf_item("general", "allow_old_protocol", CF_YESNO, conf_set_general_allow_old_protocol);
	add_conf_item("general", "verbose", CF_YESNO, conf_set_general_verbose);

	add_top_conf("channel", NULL, NULL);
	add_conf_item("channel", "use_except", CF_YESNO, conf_set_channel_use_except);
	add_conf_item("channel", "use_invex", CF_YESNO, conf_set_channel_use_invex);
	add_conf_item("channel", "use_knock", CF_YESNO, conf_set_channel_use_knock);
	add_conf_item("channel", "max_bans", CF_INT, conf_set_channel_max_bans);
	add_conf_item("channel", "knock_delay", CF_TIME, conf_set_channel_knock_delay);
	add_conf_item("channel", "knock_delay_channel", CF_TIME,
		      conf_set_channel_knock_delay_channel);
	add_conf_item("channel", "max_chans_per_user", CF_INT, conf_set_channel_max_chans_per_user);
	add_conf_item("channel", "quiet_on_ban", CF_YESNO, conf_set_channel_quiet_on_ban);
	add_conf_item("channel", "default_split_user_count", CF_INT,
		      conf_set_channel_default_split_user_count);
	add_conf_item("channel", "default_split_server_count", CF_INT,
		      conf_set_channel_default_split_server_count);
	add_conf_item("channel", "no_create_on_split", CF_YESNO,
		      conf_set_channel_no_create_on_split);
	add_conf_item("channel", "no_join_on_split", CF_YESNO, conf_set_channel_no_join_on_split);

	add_top_conf("wingate", NULL, NULL);
	add_conf_item("wingate", "wingate_enable", CF_YESNO, conf_set_wingate_enable);
	add_conf_item("wingate", "monitorbot", CF_QSTRING, conf_set_wingate_monitorbot);
	add_conf_item("wingate", "wingate_webpage", CF_QSTRING, conf_set_wingate_webpage);

	add_top_conf("hiding", NULL, NULL);
	add_conf_item("hiding", "whois_hiding", CF_QSTRING, conf_set_hiding_whois_hiding);
	add_conf_item("hiding", "whois_description", CF_QSTRING, conf_set_hiding_whois_description);
	add_conf_item("hiding", "roundrobin", CF_QSTRING, conf_set_hiding_roundrobin);
	add_conf_item("hiding", "disable_hiding", CF_YESNO, conf_set_hiding_disable_hiding);
	add_conf_item("hiding", "hidden_server", CF_YESNO, conf_set_hiding_hidden_server);
	add_conf_item("hiding", "hide_netsplit_source", CF_YESNO, conf_set_hiding_hide_netsplit_source);

	add_top_conf("badwords", NULL, NULL);
	add_conf_item("badwords", "filter", CF_QSTRING, conf_set_badwords_filter);

	add_top_conf("logging", NULL, NULL);
	add_conf_item("logging", "log_level", CF_STRING, conf_set_logging_log_level);
	add_conf_item("logging", "fuserlog", CF_QSTRING, conf_set_logging_fuserlog);
	add_conf_item("logging", "ffailed_operlog", CF_QSTRING, conf_set_logging_ffailed_operlog);
	add_conf_item("logging", "foperlog", CF_QSTRING, conf_set_logging_foperlog);
	add_conf_item("logging", "use_logging", CF_YESNO, conf_set_logging_use_logging);

	add_top_conf("alias", conf_begin_alias, conf_end_alias);
	add_conf_item("alias", "command", CF_QSTRING, conf_set_alias_command);
	add_conf_item("alias", "override", CF_YESNO, conf_set_alias_override);

}
