/*
 *  shadowircd: an advanced Internet Relay Chat Daemon(ircd).
 *  send.c: Functions for sending messages.
 *
 *  Copyright (C) 2002 by the past and present ircd coders, and others.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  $Id: send.c,v 3.6.4.17 2004/11/13 23:05:27 nenolod Exp $
 */

#include "stdinc.h"
#include "tools.h"
#include "send.h"
#include "channel.h"
#include "channel_mode.h"
#include "client.h"
#include "common.h"
#include "irc_string.h"
#include "ircd.h"
#include "handlers.h"
#include "numeric.h"
#include "fdlist.h"
#include "s_bsd.h"
#include "s_serv.h"
#include "sprintf_irc.h"
#include "s_conf.h"
#include "list.h"
#include "s_log.h"
#include "memory.h"
#include "hook.h"
#include "dh.h"
#include "s_zip.h"
#include "linebuf.h"

#define LOG_BUFSIZE 2048

static void send_message(struct Client *, char *, int);
static void send_message_remote(struct Client *, struct Client *, char *, int);

/* global for now *sigh* */
unsigned long current_serial = 0L;

/* send_format()
 *
 * inputs	- buffer to format into
 *              - size of the buffer
 *		- format pattern to use
 *		- var args
 * output	- number of bytes formatted output
 * side effects	- modifies sendbuf
 */
static inline int
send_format(char *lsendbuf, int bufsize, const char *pattern, va_list args)
{
	int len;

	/*
	 * from rfc1459
	 *
	 * IRC messages are always lines of characters terminated with a CR-LF
	 * (Carriage Return - Line Feed) pair, and these messages shall not
	 * exceed 512 characters in length,  counting all characters 
	 * including the trailing CR-LF.
	 * Thus, there are 510 characters maximum allowed
	 * for the command and its parameters.  There is no provision for
	 * continuation message lines.  See section 7 for more details about
	 * current implementations.
	 */
	len = vsnprintf(lsendbuf, bufsize - 1, pattern, args);
	if(len > bufsize - 2)
		len = bufsize - 2;	/* required by some versions of vsnprintf */

	/*
	 * We have to get a \r\n\0 onto sendbuf[] somehow to satisfy
	 * the rfc. We must assume sendbuf[] is defined to be 513
	 * bytes - a maximum of 510 characters, the CR-LF pair, and
	 * a trailing \0, as stated in the rfc. Now, if len is greater
	 * than the third-to-last slot in the buffer, an overflow will
	 * occur if we try to add three more bytes, if it has not
	 * already occured. In that case, simply set the last three
	 * bytes of the buffer to \r\n\0. Otherwise, we're ok. My goal
	 * is to get some sort of vsnprintf() function operational
	 * for this routine, so we never again have a possibility
	 * of an overflow.
	 * -wnder
	 * Exactly, vsnprintf() does the job and we don't need to check
	 * whether len > 510. We also don't need to terminate the buffer
	 * with a '\0', since the dbuf code is raw-oriented. --adx
	 */

	lsendbuf[len++] = '\r';
	lsendbuf[len++] = '\n';
	return (len);
}

/*
 ** send_message
 **      Internal utility which appends given buffer to the sockets
 **      sendq.
 */
static void
send_message(struct Client *to, char *buf, int len)
{

#ifdef INVARIANTS
	if(IsMe(to))
	{
		sendto_realops_flags(UMODE_ALL, L_ALL, "Trying to send message to myself!");
		return;
	}
#endif

	if(HasUmode(to, UMODE_NOCOLOUR))
		strip_colour(buf);

	if(linebuf_length(&to->localClient->buf_sendq) + len > get_sendq(to))
	{
		if(IsServer(to))
			sendto_realops_flags(UMODE_ALL, L_ALL,
					     "Max SendQ limit exceeded for %s: %lu > %lu",
					     get_client_name(to, HIDE_IP),
					     (unsigned
					      long) (linebuf_length(&to->localClient->buf_sendq) +
						     len), get_sendq(to));
		if(IsClient(to))
			SetSendQExceeded(to);
		dead_link_on_write(to, 0);
		return;
	}

	linebuf_parse(&to->localClient->buf_sendq, buf, len, 0);

	/*
	 ** Update statistics. The following is slightly incorrect
	 ** because it counts messages even if queued, but bytes
	 ** only really sent. Queued bytes get updated in SendQueued.
	 */
	to->localClient->sendM += 1;
	me.localClient->sendM += 1;

	if(linebuf_length(&to->localClient->buf_sendq) >
	   (IsServer(to) ? (unsigned int) 1024 : (unsigned int) 4096))
		send_queued_write(to);
}

/* send_message_remote()
 *
 * inputs	- pointer to client from message is being sent
 * 		- pointer to client to send to
 *		- pointer to preformatted buffer
 *		- length of input buffer
 * output	- none
 * side effects	- Despite the function name, this only sends to directly
 *		  connected clients.
 * 
 */
static void
send_message_remote(struct Client *to, struct Client *from, char *buf, int len)
{
	if(!MyConnect(to))
	{
		sendto_realops_flags(UMODE_ALL, L_ALL,
				     "server send message to %s [%s] dropped from %s(Not local server)",
				     to->name, to->from->name, from->name);
		return;
	}

	/* Optimize by checking if (from && to) before everything */
	/* we set to->from up there.. */

	if(!MyClient(from) && IsPerson(to) && (to == from->from))
	{
		if(IsServer(from))
		{
			sendto_realops_flags(UMODE_ALL, L_ALL,
					     "Send message to %s [%s] dropped from %s(Fake Dir)",
					     to->name, to->from->name, from->name);
			return;
		}

		sendto_realops_flags(UMODE_ALL, L_ALL,
				     "Ghosted: %s[%s@%s] from %s[%s@%s] (%s)",
				     to->name, to->username, to->host,
				     from->name, from->username, from->host, to->from->name);

		sendto_server(NULL, PROTO_CUR, PROTO_CUR,
			      ":%s KILL %s :%s (%s[%s@%s] Ghosted %s)",
			      me.id, to->name, me.name, to->name,
			      to->username, to->host, to->from->name);
		sendto_server(NULL, PROTO_NUL, PROTO_MIN,
			      ":%s KILL %s :%s (%s[%s@%s] Ghosted %s)",
			      me.name, to->name, me.name, to->name,
			      to->username, to->host, to->from->name);

		SetKilled(to);

		if(IsPerson(from))
			sendto_one(from, form_str(ERR_GHOSTEDCLIENT),
				   me.name, from->name, to->name, to->username, to->host, to->from);

		exit_client(NULL, to, &me, "Ghosted client");

		return;
	}

	send_message(to, buf, len);
}

/*
 ** sendq_unblocked
 **      Called when a socket is ready for writing.
 */
static void
sendq_unblocked(int fd, struct Client *client_p)
{
	ClearSendqBlocked(client_p);
	/* let send_queued_write be executed by send_queued_all */
}

/*
 ** slinkq_unblocked
 **      Called when a server control socket is ready for writing.
 */
static void
slinkq_unblocked(int fd, struct Client *client_p)
{
	ClearSlinkqBlocked(client_p);
	send_queued_slink_write(client_p);
}

#define FLAGS_DEBUG UMODE_DEBUG	/* I am a lazy bastard */

/*
 ** send_queued_write
 **      This is called when there is a chance that some output would
 **      be possible. This attempts to empty the send queue as far as
 **      possible, and then if any data is left, a write is rescheduled.
 */
void
send_queued_write(struct Client *to)
{
	int retlen;
#ifndef NDEBUG
	struct hook_io_data hdata;
#endif

	/*
	 ** Once socket is marked dead, we cannot start writing to it,
	 ** even if the error is removed...
	 */
	if(IsDead(to) || IsSendqBlocked(to))
		return;		/* no use calling send() now */

	/* Next, lets try to write some data */
	if(linebuf_length(&to->localClient->buf_sendq))
	{
#ifndef NDEBUG
		hdata.connection = to;
#endif
		while((retlen = linebuf_flush(to, &to->localClient->buf_sendq)) > 0)
		{
			to->localClient->sendB += retlen;
			me.localClient->sendB += retlen;
			if(to->localClient->sendB > 1023)
			{
				to->localClient->sendK += (to->localClient->sendB >> 10);
				to->localClient->sendB &= 0x03ff;	/* 2^10 = 1024, 3ff = 1023 */
			}
			else if(me.localClient->sendB > 1023)
			{
				me.localClient->sendK += (me.localClient->sendB >> 10);
				me.localClient->sendB &= 0x03ff;
			}
		}

		if((retlen < 0) && (ignoreErrno(errno)))
		{
			/* we have a non-fatal error, reschedule a write */
			SetSendqBlocked(to);
			comm_setselect(to->localClient->fd, FDLIST_IDLECLIENT, COMM_SELECT_WRITE,
				       (PF *) sendq_unblocked, (void *) to, 0);
		}

		else if(retlen <= 0)
		{
			dead_link_on_write(to, errno);
			return;
		}
	}
}

/*
 ** send_queued_slink_write
 **      This is called when there is a chance the some output would
 **      be possible. This attempts to empty the send queue as far as
 **      possible, and then if any data is left, a write is rescheduled.
 */
void
send_queued_slink_write(struct Client *to)
{
	int retlen;

	/*
	 ** Once socket is marked dead, we cannot start writing to it,
	 ** even if the error is removed...
	 */
	if(IsDead(to) || IsSlinkqBlocked(to))
		return;		/* no use calling send() now */

	/* Next, lets try to write some data */
	if(to->localClient->slinkq != NULL)
	{
		retlen = send(to->localClient->ctrlfd,
			      to->localClient->slinkq + to->localClient->slinkq_ofs,
			      to->localClient->slinkq_len, 0);
		if(retlen < 0)
		{
			/* If we have a fatal error */
			if(!ignoreErrno(errno))
			{
				dead_link_on_write(to, errno);
				return;
			}
		}
		else if(retlen == 0)
		{
			/* 0 bytes is an EOF .. */
			dead_link_on_write(to, 0);
			return;
		}
		else
		{
			to->localClient->slinkq_len -= retlen;

			assert(to->localClient->slinkq_len >= 0);
			if(to->localClient->slinkq_len)
				to->localClient->slinkq_ofs += retlen;
			else
			{
				to->localClient->slinkq_ofs = 0;
				MyFree(to->localClient->slinkq);
				to->localClient->slinkq = NULL;
			}
		}

		/* Finally, if we have any more data, reschedule a write */
		if(to->localClient->slinkq_len)
		{
			SetSlinkqBlocked(to);
			comm_setselect(to->localClient->ctrlfd, FDLIST_IDLECLIENT,
				       COMM_SELECT_WRITE, (PF *) slinkq_unblocked, (void *) to, 0);
		}
	}
}

/* send_queued_all()
 *
 * input        - NONE
 * output       - NONE
 * side effects - try to flush sendq of each client
 */
void
send_queued_all(void)
{
	dlink_node *ptr;

	/* Servers are processed first, mainly because this can generate
	 * a notice to opers, which is to be delivered by this function.
	 */
	DLINK_FOREACH(ptr, serv_list.head) send_queued_write((struct Client *) ptr->data);

	DLINK_FOREACH(ptr, unknown_list.head) send_queued_write((struct Client *) ptr->data);

	DLINK_FOREACH(ptr, local_client_list.head) send_queued_write((struct Client *) ptr->data);

	/* NOTE: This can still put clients on aborted_list; unfortunately,
	 * exit_aborted_clients takes precedence over send_queued_all,
	 * because exiting clients can generate server/oper traffic.
	 * The easiest approach is dealing with aborted clients in the next I/O lap.
	 * -adx
	 */
}

/* sendto_one()
 *
 * inputs	- pointer to destination client
 *		- var args message
 * output	- NONE
 * side effects	- send message to single client
 */
void
sendto_one(struct Client *to, const char *pattern, ...)
{
	va_list args;
	char buffer[IRCD_BUFSIZE];
	int len;

	if(to->from != NULL)
		to = to->from;
	if(IsDead(to))
		return;		/* This socket has already been marked as dead */

	va_start(args, pattern);
	len = send_format(buffer, IRCD_BUFSIZE, pattern, args);
	va_end(args);

	send_message(to, buffer, len);
}

/* sendto_one_prefix()
 *
 * inputs	- pointer to destination client
 *		- pointer to client to form prefix from
 *		- var args message
 * output	- NONE
 * side effects	- send message to single client
 */
void
sendto_one_prefix(struct Client *to, struct Client *prefix, const char *pattern, ...)
{
	va_list args;
	char buffer[IRCD_BUFSIZE];
	int len;

	if(to->from != NULL)
		to = to->from;
	if(IsDead(to))
		return;		/* This socket has already been marked as dead */

	len = ircsprintf(buffer, ":%s ", (IsServer(to) && (to->protocol > PROTO_MIN) ?
					  ID(prefix) : prefix->name));

	va_start(args, pattern);
	len += send_format(&buffer[len], IRCD_BUFSIZE - len, pattern, args);
	va_end(args);

	send_message(to, buffer, len);
}

/* sendto_channel_butone()
 *
 * inputs	- pointer to client(server) to NOT send message to
 *		- pointer to client that is sending this message
 *		- pointer to channel being sent to
 *		- vargs message
 * output	- NONE
 * side effects	- message as given is sent to given channel members.
 */
void
sendto_channel_butone(struct Client *one, struct Client *from,
		      struct Channel *chptr, const char *command, const char *pattern, ...)
{
	va_list args;
	char local_buf[IRCD_BUFSIZE];
	char remote_buf[IRCD_BUFSIZE];
	char uid_buf[IRCD_BUFSIZE];
	int local_len, remote_len, uid_len;
	dlink_node *ptr;
	dlink_node *ptr_next;
	struct Client *target_p;

	chptr->last_privmsg = time(NULL);

	if(IsServer(from))
		local_len = ircsprintf(local_buf, ":%s %s %s ", from->name, command, chptr->chname);
	else
		if (chptr->mode.mode & MODE_ANONYMITY)
			local_len = ircsprintf(local_buf, ":...!...@... %s %s ",
						command, chptr->chname);
		else
			local_len = ircsprintf(local_buf, ":%s!%s@%s %s %s ",
					       from->name, from->username, GET_CLIENT_HOST(from),
					       command, chptr->chname);

	remote_len = ircsprintf(remote_buf, ":%s %s %s ", from->name, command, chptr->chname);
	uid_len = ircsprintf(uid_buf, ":%s %s %s ", ID(from), command, chptr->chname);

	va_start(args, pattern);
	local_len += send_format(&local_buf[local_len], IRCD_BUFSIZE - local_len, pattern, args);
	remote_len += send_format(&remote_buf[remote_len], IRCD_BUFSIZE - remote_len,
				  pattern, args);
	uid_len += send_format(&uid_buf[uid_len], IRCD_BUFSIZE - uid_len, pattern, args);
	va_end(args);

	++current_serial;

	DLINK_FOREACH_SAFE(ptr, ptr_next, chptr->members.head)
	{
		target_p = ((struct Membership *) ptr->data)->client_p;
		assert(target_p != NULL);

		if(IsDefunct(target_p) || target_p->from == one)
			continue;

		if(MyConnect(target_p) && IsRegisteredUser(target_p))
		{
			if(target_p->serial != current_serial)
			{
				send_message(target_p, local_buf, local_len);
				target_p->serial = current_serial;
			}
		}
		else
		{
			/* Now check whether a message has been sent to this
			 * remote link already
			 */
			if(target_p->from->serial != current_serial)
			{
				if(target_p->from->protocol > PROTO_MIN)
					send_message_remote(target_p->from, from, uid_buf, uid_len);
				else
					send_message_remote(target_p->from, from, remote_buf,
							    remote_len);
				target_p->from->serial = current_serial;
			}
		}
	}
}

/* sendto_server()
 * 
 * inputs       - pointer to client to NOT send to
 *              - caps or'd together which must ALL be present
 *              - caps or'd together which must ALL NOT be present
 *              - printf style format string
 *              - args to format string
 * output       - NONE
 * side effects - Send a message to all connected servers, which match
 *                the required protocol commands.
 */
void
sendto_server(struct Client *one, int protomin, int protomax, const char *format, ...)
{
	va_list args;
	struct Client *client_p;
	dlink_node *ptr;
	char buffer[IRCD_BUFSIZE];
	int len;

	va_start(args, format);
	len = send_format(buffer, IRCD_BUFSIZE, format, args);
	va_end(args);

	DLINK_FOREACH(ptr, serv_list.head)
	{
		client_p = ptr->data;

		/* If dead already skip */
		if(IsDead(client_p))
			continue;

		/* check against 'one' */
		if(one != NULL && (client_p == one->from))
			continue;

		/* check that we have a valid protocol version */
		if(client_p->protocol < protomin || client_p->protocol > protomax)
			continue;

		send_message(client_p, buffer, len);
	}
}

/* sendto_common_channels_local()
 *
 * inputs	- pointer to client
 *		- pattern to send
 * output	- NONE
 * side effects	- Sends a message to all people on local server who are
 * 		  in same channel with user. 
 *		  used by m_nick.c and exit_one_client.
 */
void
sendto_common_channels_local(struct Client *user, int touser, const char *pattern, ...)
{
	va_list args;
	dlink_node *uptr;
	dlink_node *cptr;
	struct Channel *chptr;
	struct Membership *ms;
	struct Client *target_p;
	char buffer[IRCD_BUFSIZE];
	int len;

	va_start(args, pattern);
	len = send_format(buffer, IRCD_BUFSIZE, pattern, args);
	va_end(args);

	++current_serial;

	DLINK_FOREACH(cptr, user->user->channel.head)
	{
		chptr = ((struct Membership *) cptr->data)->chptr;
		assert(chptr != NULL);

		DLINK_FOREACH(uptr, chptr->locmembers.head)
		{
			ms = uptr->data;
			target_p = ms->client_p;
			assert(target_p != NULL);

			if(target_p == user || IsDefunct(target_p) ||
			   target_p->serial == current_serial)
				continue;

			target_p->serial = current_serial;
			send_message(target_p, buffer, len);
		}
	}

	if(touser && MyConnect(user) && !IsDead(user) && user->serial != current_serial)
		send_message(user, buffer, len);
}

/* sendto_channel_local()
 *
 * inputs	- member status mask, e.g. CHFL_CHANOP | CHFL_VOICE
 *              - pointer to channel to send to
 *              - var args pattern
 * output	- NONE
 * side effects - Send a message to all members of a channel that are
 *		  locally connected to this server.
 */
void
sendto_channel_local(int type, struct Channel *chptr, const char *pattern, ...)
{
	va_list args;
	char buffer[IRCD_BUFSIZE];
	int len;
	dlink_node *ptr;
	struct Membership *ms;
	struct Client *target_p;

	chptr->last_privmsg = time(NULL);

	va_start(args, pattern);
	len = send_format(buffer, IRCD_BUFSIZE, pattern, args);
	va_end(args);

	/* Serial number checking isn't strictly necessary, but won't hurt */
	++current_serial;

	DLINK_FOREACH(ptr, chptr->locmembers.head)
	{
		ms = ptr->data;
		target_p = ms->client_p;

		if(type != 0 && (ms->flags & type) == 0)
			continue;

		if(IsDefunct(target_p) || target_p->serial == current_serial)
			continue;

		if(HasUmode(target_p, UMODE_DEAF))
			continue;

		target_p->serial = current_serial;

		send_message(target_p, buffer, len);
	}
}

/* sendto_channel_local_butone()
 *
 * inputs       - pointer to client to NOT send message to
 *              - member status mask, e.g. CHFL_CHANOP | CHFL_VOICE
 *              - pointer to channel to send to
 *              - var args pattern
 * output       - NONE
 * side effects - Send a message to all members of a channel that are
 *                locally connected to this server except one.
 */
void
sendto_channel_local_butone(struct Client *one, int type,
			    struct Channel *chptr, const char *pattern, ...)
{
	va_list args;
	char buffer[IRCD_BUFSIZE];
	int len;
	struct Client *target_p;
	struct Membership *ms;
	dlink_node *ptr;

	chptr->last_privmsg = time(NULL);

	va_start(args, pattern);
	len = send_format(buffer, IRCD_BUFSIZE, pattern, args);
	va_end(args);

	/* Serial number checking isn't strictly necessary, but won't hurt */
	++current_serial;

	DLINK_FOREACH(ptr, chptr->locmembers.head)
	{
		ms = ptr->data;
		target_p = ms->client_p;

		if(type != 0 && (ms->flags & type) == 0)
			continue;

		if(target_p == one || IsDefunct(target_p) || target_p->serial == current_serial)
			continue;

		target_p->serial = current_serial;

		send_message(target_p, buffer, len);
	}
}


/* sendto_channel_remote()
 *
 * inputs	- Client not to send towards
 *		- Client from whom message is from
 *		- member status mask, e.g. CHFL_CHANOP | CHFL_VOICE
 *              - pointer to channel to send to
 *              - var args pattern
 * output	- NONE
 * side effects - Send a message to all members of a channel that are
 *		  remote to this server.
 */
void
sendto_channel_remote(struct Client *one, struct Client *from, int type, struct Channel *chptr,
		      const char *pattern, ...)
{
	va_list args;
	char buffer[IRCD_BUFSIZE];
	int len;
	dlink_node *ptr;
	struct Client *target_p;
	struct Membership *ms;

	chptr->last_privmsg = time(NULL);

	va_start(args, pattern);
	len = send_format(buffer, IRCD_BUFSIZE, pattern, args);
	va_end(args);

	/* Serial number checking isn't strictly necessary, but won't hurt */
	++current_serial;

	DLINK_FOREACH(ptr, chptr->members.head)
	{
		ms = ptr->data;
		target_p = ms->client_p;

		if(type != 0 && (ms->flags & type) == 0)
			continue;

		if(MyConnect(target_p))
			continue;
		target_p = target_p->from;

		if(target_p == one->from || target_p->from->serial == current_serial)
			continue;

		target_p->serial = current_serial;

		send_message(target_p, buffer, len);
	}
}

/*
 ** match_it() and sendto_match_butone() ARE only used
 ** to send a msg to all ppl on servers/hosts that match a specified mask
 ** (used for enhanced PRIVMSGs) for opers
 **
 ** addition -- Armin, 8jun90 (gruner@informatik.tu-muenchen.de)
 **
 */

/* match_it()
 *
 * inputs	- client pointer to match on
 *		- actual mask to match
 *		- what to match on, HOST or SERVER
 * output	- 1 or 0 if match or not
 * side effects	- NONE
 */
static int
match_it(const struct Client *one, const char *mask, int what)
{
	if(what == MATCH_HOST)
		return (match(mask, one->host));

	return (match(mask, one->user->server->name));
}

/* sendto_match_butone()
 *
 * Send to all clients which match the mask in a way defined on 'what';
 * either by user hostname or user servername.
 *
 * ugh. ONLY used by m_message.c to send an "oper magic" message. ugh.
 */
void
sendto_match_butone(struct Client *one, struct Client *from, char *mask,
		    int what, const char *pattern, ...)
{
	va_list args;
	struct Client *client_p;
	dlink_node *ptr, *ptr_next;
	char local_buf[IRCD_BUFSIZE], remote_buf[IRCD_BUFSIZE];
	int local_len = ircsprintf(local_buf, ":%s!%s@%s ", from->name,
				   from->username, GET_CLIENT_HOST(from));
	int remote_len = ircsprintf(remote_buf, ":%s ", from->name);

	va_start(args, pattern);
	local_len += send_format(&local_buf[local_len], IRCD_BUFSIZE - local_len, pattern, args);
	remote_len += send_format(&remote_buf[remote_len], IRCD_BUFSIZE - remote_len,
				  pattern, args);
	va_end(args);

	/* scan the local clients */
	DLINK_FOREACH(ptr, local_client_list.head)
	{
		client_p = ptr->data;

		if(client_p != one && !IsDefunct(client_p) && match_it(client_p, mask, what))
			send_message(client_p, local_buf, local_len);
	}

	/* Now scan servers */
	DLINK_FOREACH_SAFE(ptr, ptr_next, serv_list.head)
	{
		client_p = ptr->data;

		/*
		 * The old code looped through every client on the
		 * network for each server to check if the
		 * server (client_p) has at least 1 client matching
		 * the mask, using something like:
		 *
		 * for (target_p = GlobalClientList; target_p; target_p = target_p->next)
		 *        if (IsRegisteredUser(target_p) &&
		 *                        match_it(target_p, mask, what) &&
		 *                        (target_p->from == client_p))
		 *   vsendto_prefix_one(client_p, from, pattern, args);
		 *
		 * That way, we wouldn't send the message to
		 * a server who didn't have a matching client.
		 * However, on a network such as EFNet, that
		 * code would have looped through about 50
		 * servers, and in each loop, loop through
		 * about 50k clients as well, calling match()
		 * in each nested loop. That is a very bad
		 * thing cpu wise - just send the message
		 * to every connected server and let that
		 * server deal with it.
		 * -wnder
		 */
		if(client_p != one && !IsDefunct(client_p))
			send_message_remote(client_p, from, remote_buf, remote_len);
	}
}

/* sendto_match_servs()
 *
 * inputs       - source client
 *              - mask to send to
 *              - data
 * outputs	- none
 * side effects	- data sent to servers matching with capab
 */
void
sendto_match_servs(struct Client *source_p, const char *mask, const char *pattern, ...)
{
	va_list args;
	struct Client *target_p;
	dlink_node *ptr;
	char buffer[BUFSIZE];
	int found = 0;

	va_start(args, pattern);
	vsnprintf((char *) &buffer, sizeof(buffer), pattern, args);
	va_end(args);

	current_serial++;

	DLINK_FOREACH(ptr, global_serv_list.head)
	{
		target_p = ptr->data;

		/* Do not attempt to send to ourselves, or the source */
		if(IsMe(target_p) || target_p->from == source_p->from)
			continue;

		if(target_p->from->serial == current_serial)
			continue;

		if(match(mask, target_p->name))
		{
			target_p->from->serial = current_serial;
			found++;

			sendto_anywhere(target_p, source_p, buffer);
		}
	}
}

/* sendto_anywhere()
 *
 * inputs	- pointer to dest client
 * 		- pointer to from client
 * 		- varags
 * output	- NONE
 * side effects	- less efficient than sendto_remote and sendto_one
 * 		  but useful when one does not know where target "lives"
 */
void
sendto_anywhere(struct Client *to, struct Client *from, const char *pattern, ...)
{
	va_list args;
	char buffer[IRCD_BUFSIZE];
	int len;
	struct Client *send_to = (to->from != NULL ? to->from : to);

	if(IsDead(send_to))
		return;

	if(MyClient(to))
	{
		if(IsServer(from))
		{
			if(to->protocol == PROTO_CUR && HasID(from))
				len = ircsprintf(buffer, ":%s ", from->id);
			else
				len = ircsprintf(buffer, ":%s ", from->name);
		}
		else
			len = ircsprintf(buffer, ":%s!%s@%s ",
					 from->name, from->username, GET_CLIENT_HOST(from));
	}
	else
		len = ircsprintf(buffer, ":%s ", ID_or_name(from, send_to));

	va_start(args, pattern);
	len += send_format(&buffer[len], IRCD_BUFSIZE - len, pattern, args);
	va_end(args);

	if(HasUmode(to, UMODE_DEAF))
		return;

	if(MyClient(to))
		send_message(send_to, buffer, len);
	else
		send_message_remote(send_to, from, buffer, len);
}

/* sendto_realops_flags()
 *
 * inputs	- flag types of messages to show to real opers
 *		- flag indicating opers/admins
 *		- var args input message
 * output	- NONE
 * side effects	- Send to *local* ops only but NOT +s nonopers.
 */
void
sendto_realops_flags(unsigned int flags, int level, const char *pattern, ...)
{
	struct Client *client_p;
	char nbuf[IRCD_BUFSIZE];
	dlink_node *ptr;
	va_list args;

	va_start(args, pattern);
	vsnprintf(nbuf, IRCD_BUFSIZE, pattern, args);
	va_end(args);

	DLINK_FOREACH(ptr, oper_list.head)
	{
		client_p = ptr->data;
		assert(HasUmode(client_p, UMODE_OPER));

		/* If we're sending it to opers and theyre an admin, skip.
		 * If we're sending it to admins, and theyre not, skip.
		 */
		if(((level == L_ADMIN) && !IsAdmin(client_p)) ||
		   ((level == L_OPER) && IsAdmin(client_p)))
			continue;

		if(HasUmode(client_p, flags))
			sendto_one(client_p, ":%s NOTICE %s :*** Notice -- %s",
				   me.name, client_p->name, nbuf);
	}
}

/* sendto_wallops_flags()
 *
 * inputs       - flag types of messages to show to real opers
 *              - client sending request
 *              - var args input message
 * output       - NONE
 * side effects - Send a wallops to local opers
 */
void
sendto_wallops_flags(unsigned int flags, struct Client *source_p, const char *pattern, ...)
{
	struct Client *client_p;
	dlink_node *ptr;
	va_list args;
	char buffer[IRCD_BUFSIZE];
	int len;

	if(IsPerson(source_p))
		len = ircsprintf(buffer, ":%s!%s@%s WALLOPS :",
				 source_p->name, source_p->username, GET_CLIENT_HOST(source_p));
	else
		len = ircsprintf(buffer, ":%s WALLOPS :", source_p->name);

	va_start(args, pattern);
	len += send_format(&buffer[len], IRCD_BUFSIZE - len, pattern, args);
	va_end(args);

	DLINK_FOREACH(ptr, local_client_list.head)
	{
		client_p = ptr->data;

		if((HasUmode(client_p, flags)) && !IsDefunct(client_p))
			send_message(client_p, buffer, len);
	}
}

/* ts_warn()
 *
 * inputs	- var args message
 * output	- NONE
 * side effects	- Call sendto_realops_flags, with some flood checking
 *		  (at most 5 warnings every 5 seconds)
 */
void
ts_warn(const char *pattern, ...)
{
	va_list args;
	char buffer[LOG_BUFSIZE];
	static time_t last = 0;
	static int warnings = 0;

	/*
	 ** if we're running with TS_WARNINGS enabled and someone does
	 ** something silly like (remotely) connecting a nonTS server,
	 ** we'll get a ton of warnings, so we make sure we don't send
	 ** more than 5 every 5 seconds.  -orabidoo
	 */

	if(CurrentTime - last < 5)
	{
		if(++warnings > 5)
			return;
	}
	else
	{
		last = CurrentTime;
		warnings = 0;
	}

	va_start(args, pattern);
	vsprintf_irc(buffer, pattern, args);
	va_end(args);

	sendto_realops_flags(UMODE_ALL, L_ALL, "%s", buffer);
	ilog(L_CRIT, "%s", buffer);
}

/* kill_client()
 *
 * inputs	- client to send kill towards
 * 		- pointer to client to kill
 * 		- reason for kill
 * output	- NONE
 * side effects	- NONE
 */
void
kill_client(struct Client *client_p, struct Client *diedie, const char *pattern, ...)
{
	va_list args;
	char buffer[IRCD_BUFSIZE];
	int len;

	if(client_p->from != NULL)
		client_p = client_p->from;
	if(IsDead(client_p))
		return;

	len = ircsprintf(buffer, ":%s KILL %s :", me.name, ID_or_name(diedie, client_p));

	va_start(args, pattern);
	len += send_format(&buffer[len], IRCD_BUFSIZE - len, pattern, args);
	va_end(args);

	send_message(client_p, buffer, len);
}
